"""Base for other SCRAPI classes."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import re

from typing import Dict
from google.oauth2 import service_account
from google.auth.transport.requests import Request
from google.protobuf import json_format  # type: ignore

from proto.marshal.collections import repeated
from proto.marshal.collections import maps


class ScrapiBase:
    """Core Class for managing Auth and other shared functions."""

    global_scopes = [
        "https://www.googleapis.com/auth/cloud-platform",
        "https://www.googleapis.com/auth/dialogflow",
    ]

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict[str, str] = None,
        creds: service_account.Credentials = None,
        scope=False,
        agent_id=None,
    ):

        self.scopes = ScrapiBase.global_scopes
        if scope:
            self.scopes += scope

        if creds:
            self.creds = creds
            self.creds.refresh(Request())
            self.token = self.creds.token
        elif creds_path:
            self.creds = service_account.Credentials.from_service_account_file(
                creds_path, scopes=self.scopes
            )
            self.creds.refresh(Request())
            self.token = self.creds.token
        elif creds_dict:
            self.creds = service_account.Credentials.from_service_account_info(
                creds_dict, scopes=self.scopes
            )
            self.creds.refresh(Request())
            self.token = self.creds.token
        else:
            self.creds = None
            self.token = None

        if agent_id:
            self.agent_id = agent_id

    @staticmethod
    def _set_region(item_id):
        """Different regions have different API endpoints

        Args:
          item_id: agent/flow/page - any type of long path id like
            `projects/<GCP PROJECT ID>/locations/<LOCATION ID>

        Returns:
          A dictionary containing the api_endpoint to use when
          instantiating other library client objects, or None
          if the location is "global"
        """
        try:
            location = item_id.split("/")[3]
        except IndexError as err:
            logging.error("IndexError - path too short? %s", item_id)
            raise err

        if location != "global":
            api_endpoint = f"{location}-dialogflow.googleapis.com:443"
            client_options = {"api_endpoint": api_endpoint}
            return client_options

        else:
            return None  # explicit None return when not required

    @staticmethod
    def pbuf_to_dict(pbuf):
        """Extractor of json from a protobuf"""
        blobstr = json_format.MessageToJson(
            pbuf
        )  # i think this returns JSON as a string
        blob = json.loads(blobstr)
        return blob

    @staticmethod
    def cx_object_to_json(cx_object):
        """Response objects have a magical _pb field attached"""
        return ScrapiBase.pbuf_to_dict(cx_object._pb)  # pylint: disable=W0212

    @staticmethod
    def cx_object_to_dict(cx_object):
        """Response objects have a magical _pb field attached"""
        return ScrapiBase.pbuf_to_dict(cx_object._pb)  # pylint: disable=W0212

    @staticmethod
    def extract_payload(msg):
        """Convert to json so we can get at the object"""
        blob = ScrapiBase.cx_object_to_dict(msg)
        return blob.get("payload")  # deref for nesting

    @staticmethod
    def _parse_resource_path(
        resource_type,
        resource_id,
        validate=True) -> Dict[str, str]:
        # pylint: disable=line-too-long
        """Validates the provided Resource ID against known patterns.

        Args:
          resource_type, Must be one of the following resource types:
            `agent`, `entity_type`, `environmnet`, `flow`, `intent`, `page`,
            `project`, `security_setting`, `session`, `session_entity_type`,
            `test_case`, `transition_route_group`, `version`, `webhook`
          resource_id, The CX resource ID to check against the provided
            resource_type
          validate, allows the user to have their Resource ID validated along
            with returning the parts dictionary of the Resource ID. If set to
            True, this method will prompt the user with the correct format
            to utilize for the specified ID. If set to False, no validation
            will occur. If the input Resource ID is invalid when set to False,
            an empty Dictionary will be returned, allowing the caller to
            define their own ValueError message in a higher level class.
            Defaults to True.
        """

        standard_id_match = r"[-0-9a-f]{1,36}"
        entity_id_match = r"[-@.0-9a-z]{1,36}"
        location_id_match = r"[-0-9a-z]{1,36}"
        session_id_match = r"[-0-9a-zA-Z!@#$%^&*()_+={}[\]:;\"'<>,.?]{1,36}"
        version_id_match = r"[0-9]{1,4}"

        matcher_root = f"^projects/(?P<project>.+?)/locations/(?P<location>{location_id_match})"

        pattern_map = {
            "agent": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`",
            },
            "entity_type": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/entityTypes/(?P<entity>{entity_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/entityTypes/<Entity Types ID>`",
            },
            "environment": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/environments/(?P<environment>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/environments/<Environment ID>`",
            },
            "flow": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>`",
            },
            "intent": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/intents/(?P<intent>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/intents/<Intent ID>`",
            },
            "page": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})/pages/(?P<page>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>/pages/<Page ID>`",
            },
            "project": {
                "matcher": fr"{matcher_root}$",
                "format": "`projects/<Project ID>/locations/<Location ID>/`",
            },
            "security_setting": {
                "matcher": fr"{matcher_root}/securitySettings/(?P<security_setting>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/securitySettings/<Security Setting ID>`",
            },
            "session": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/sessions/(?P<session>{session_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/sessions/<Session ID>`",
            },
            "session_entity_type": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/sessions/(?P<session>{session_id_match})/entityTypes/(?P<entity>{entity_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/sessions/<Session ID>/entityTypes/<Entity Type ID>`",
            },
            "test_case": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/testCases/(?P<test_case>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/testCases/<Test Case ID>`",
            },
            "transition_route_group": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})/transitionRouteGroups/(?P<transition_route_group>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>/transitionRouteGroups/<Transition Route Group ID>`",
            },
            "version": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})/versions/(?P<version>{version_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>`",
            },
            "webhook": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/webhooks/(?P<webhook>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`",
            }
        }

        if resource_type not in pattern_map:
            raise KeyError(
                "`resource_type` must be one of the following resource types:"
                " `agent`, `entity_type`, `environmnet`, `flow`, `intent`,"
                "`page`,`project`, `security_setting`, `session`, "
                "`session_entity_type`,`test_case`, `transition_route_group`, "
                "`version`, `webhook`"
            )

        match_res = re.match(pattern_map[resource_type]["matcher"], resource_id)
        dict_res = match_res.groupdict() if match_res else {}
        valid_parse = False

        if dict_res:
            valid_parse = True

        if validate and not valid_parse:
            raise ValueError(
                f"{resource_type.capitalize()} ID must be provided in the "
                f"following format: "
                f"{pattern_map[resource_type]['format']}"
            )

        # pylint: enable=line-too-long
        return dict_res

    def recurse_proto_repeated_composite(self, repeated_object):
        repeated_list = []
        for item in repeated_object:
            if isinstance(item, repeated.RepeatedComposite):
                item = self.recurse_proto_repeated_composite(item)
                repeated_list.append(item)
            elif isinstance(item, maps.MapComposite):
                item = self.recurse_proto_marshal_to_dict(item)
                repeated_list.append(item)
            else:
                repeated_list.append(item)

        return repeated_list

    def recurse_proto_marshal_to_dict(self, marshal_object):
        new_dict = {}
        for k, v in marshal_object.items():
            if isinstance(v, maps.MapComposite):
                v = self.recurse_proto_marshal_to_dict(v)
            elif isinstance(v, repeated.RepeatedComposite):
                v = self.recurse_proto_repeated_composite(v)
            new_dict[k] = v

        return new_dict



"""Unit Tests for Search Util Class"""
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from src.dfcx_scrapi.tools import search_util

@pytest.mark.unit
def test_instantiate_search_util(creds):
    scrapi_search = search_util.SearchUtil(creds_path=creds)

    assert isinstance(scrapi_search, search_util.SearchUtil)
    assert scrapi_search.creds_path == creds

@pytest.mark.unit
def test_get_agent_fulfillments(creds, agent_id):
    scrapi_search = search_util.SearchUtil(creds_path=creds)
    df = scrapi_search.get_agent_fulfillment_message_df(agent_id)
    assert set(df.columns) == {
            'flow_name',
            'page_name',
            'parameter_name',
            'event',
            'route_group_name',
            'intent',
            'condition',
            'response_message',
            'conditional_cases',
            'response_type'}



"""CCAI Security Settings Methods"""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
import json
import logging
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2
from dfcx_scrapi.core import scrapi_base


# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class SecuritySettings(scrapi_base.ScrapiBase):
    """Core Class for CCAI Security Settings."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict[str,str] = None,
        creds=None,
        scope=False,
        agent_id: str = None
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if agent_id:
            self.agent_id = agent_id

        self.ss_service = services.security_settings_service
        self.ss_types = types.security_settings

    def list_security_settings(self, location_id: str):
        """List Security Settings for a given Project and Region.

        Args:
          location_id: CX Agent ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/

        Returns:
          List of security settings
        """

        request = self.ss_types.ListSecuritySettingsRequest()

        request.parent = location_id
        client_options = self._set_region(location_id)
        client = self.ss_service.SecuritySettingsServiceClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_security_settings(request)

        security_settings = []
        for page in response.pages:
            for ss in page.security_settings:
                security_settings.append(ss)

        return security_settings


    def get_security_settings(self, security_setting_id: str):
        """Get specified CCAI Security Setting.

        Args:
          security_setting_id: Security Setting ID string in the following
            format: projects/<PROJECT ID>/locations/<LOCATION ID>/
            securitySettings/<SECURITY SETTINGS ID>

        Returns:
          A single Security Settings object of types.SecuritySettings
        """

        request = self.ss_types.GetSecuritySettingsRequest()

        request.name = security_setting_id
        client_options = self._set_region(security_setting_id)
        client = self.ss_service.SecuritySettingsServiceClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_security_settings(request)

        return response

    def create_security_settings(
        self,
        location_id: str,
        obj: types.SecuritySettings = None,
        security_settings_dict: Dict[str,str] = None):
        """Create CCAI Security Settings profile in the specified project.

        One of `obj` or `security_settings_dict` should be provided to create a
        new security settings profile.

        Args:
          location_id: CX Agent ID string in the following format:
            projects/<PROJECT ID>/locations/<LOCATION ID>/
          obj: An object of types.SecuritySettings representing the Security
            Settings object to be created
          security_settings_dict: An optional dictionary of key/value pairs
            that correspond to the fields and values necessary for creating a
            new Security Settings profile

        Returns:
          A single Security Settings object of types.SecuritySettings
        """

        if obj and security_settings_dict:
            raise ValueError(
                "Cannot provide both obj and security_settings_dict")
        elif obj:
            security_settings = obj
            security_settings.name = ""
        elif security_settings_dict:
            security_settings = self.ss_types.SecuritySettings.from_json(
                json.dumps(security_settings_dict)
            )
        else:
            raise ValueError(
                "Must provide either obj or security_settings_dict")

        request = self.ss_types.CreateSecuritySettingsRequest()

        request.parent = location_id
        request.security_settings = security_settings

        client_options = self._set_region(location_id)
        client = self.ss_service.SecuritySettingsServiceClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.create_security_settings(request)

        return response

    def update_security_settings(self, security_setting_id: str, **kwargs):
        """Update specified CCAI Security Setting.

        Args:
          security_setting_id: Security Setting ID string in the following
            format: projects/<PROJECT ID>/locations/<LOCATION ID>/
            securitySettings/<SECURITY SETTINGS ID>

        Returns:
          A single Security Settings object of types.SecuritySettings
        """

        security_settings = self.get_security_settings(security_setting_id)

        # set intent attributes from kwargs
        for key, value in kwargs.items():
            setattr(security_settings, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        request = self.ss_types.UpdateSecuritySettingsRequest()

        request.security_settings = security_settings
        request.update_mask = mask

        client_options = self._set_region(security_setting_id)
        client = self.ss_service.SecuritySettingsServiceClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.update_security_settings(request)

        return response

    def delete_security_settings(self, security_setting_id: str):
        """Delete the specified CCAI Security Setting.

        Args:
          security_setting_id: Security Setting ID string in the following
            format: projects/<PROJECT ID>/locations/<LOCATION ID>/
            securitySettings/<SECURITY SETTINGS ID>
        """
        request = self.ss_types.DeleteSecuritySettingsRequest()
        request.name = security_setting_id

        client_options = self._set_region(security_setting_id)
        client = self.ss_service.SecuritySettingsServiceClient(
            credentials=self.creds, client_options=client_options
        )

        client.delete_security_settings(request)



"""Session Entity Types Resource functions."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List

from google.oauth2 import service_account
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class SessionEntityTypes(scrapi_base.ScrapiBase):
    """Core Class for CX Session Entity Type functions.
    https://cloud.google.com/dialogflow/cx/docs/concept/entity-session
    """

    def init(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds: service_account.Credentials = None,
        scope: bool = False,
        agent_id: str = None,
    ):

        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id

    def _merge_session_id_and_env_id(
        self, session_id: str, environment_id: str
    ) -> str:
        """Merges the Session ID and Environment ID into a valid Session ID.

        When using DFCX Environments, the Session ID needs to be modified to
        include the Environment ID as part of its fully qualified path. Use
        this method to merge the 2 IDs into a valid Session ID when a non-DRAFT
        environment is in use.

        session_id, The session to list all session entity types from.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              sessions/<Session ID>``
          environment_id, The Environment associated with the Session ID to
            list all session entity types from. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              environments/<Environment ID>`
        """
        env_parts = self._parse_resource_path("environment", environment_id)
        session_parts = self._parse_resource_path("session", session_id)

        parent_id = (
            f"projects/{session_parts['project']}/locations/"
            f"{session_parts['location']}/agents/{session_parts['agent']}"
            f"/environments/{env_parts['environment']}/sessions/"
            f"{session_parts['session']}"
        )

        return parent_id

    def _merge_session_entity_id_and_env_id(
        self, session_entity_type_id: str, environment_id: str
    ) -> str:
        """Merges the Session Entity Type ID and Environment ID.

        When using DFCX Environments, the Session Entity Type ID needs to be
        modified to include the Environment ID as part of its fully qualified
        path. Use this method to merge the 2 IDs into a valid Session Entity
        Type ID when a non-DRAFT environment is in use.

        Args:
          session_entity_type_id, The Session Entity Type ID to merge.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              sessions/<Session ID>/entityTypes/<Entity Type ID>``
          environment_id, The Environment associated with the Session Entity
            Type ID to merge. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              environments/<Environment ID>`
        """
        entity_parts = self._parse_resource_path(
            "session_entity_type", session_entity_type_id
        )
        environment_parts = self._parse_resource_path(
            "environment", environment_id
        )

        parent_id = (
            f"projects/{environment_parts['project']}/"
            f"locations/{environment_parts['location']}/"
            f"agents/{environment_parts['agent']}/"
            f"environments/{environment_parts['environment']}/"
            f"sessions/{environment_parts['session']}/"
            f"entityTypes/{entity_parts['entity']}"
        )

        return parent_id

    def build_session_entity_type(
        self,
        session_id: str,
        entity_id: str,
        entity_dict: Dict[str, List[str]],
        environment_id: str = None,
        entity_override_mode: str = "Override",
    ) -> types.SessionEntityType:
        """Builds a Session Entity Type object based on simple inputs.

        Args:
          session_id, The session to list all session entity types from.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              sessions/<Session ID>``
          entity_id, The Entity ID to replace or extend that currently exists
            in the DFCX Agent.
          entity_dict, The Values and associated Synonyms for the Session
          Entities to be created.
            Ex:
              [1] {'scallions': ['green onions']}
              [2] {'fruit': ['apple','orange'],
                'vegetables': ['cabbage','celery']}
          environment_id, The Environment associated with the Session ID to
            list all session entity types from. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              environments/<Environment ID>`
            If ``Environment ID`` is not specified, we assume default 'DRAFT'
              environment.
          entity_override_mode, ONEOF: `Override`, `Supplement`
            `Override` will override the current collection of Session Entity
              Types for the corresponding entity type.
            `Supplement` will extend the current collection of Session Entity
              Types for the corresponding entity type.
        """
        if environment_id:
            parent_id = self._merge_session_id_and_env_id(
                session_id, environment_id
            )

        else:
            parent_id = session_id

        # Append Entity ID to Parent ID
        entity_parts = self._parse_resource_path("entity_type", entity_id)
        parent_id += f"/entityTypes/{entity_parts['entity']}"

        st = types.SessionEntityType()
        st.name = parent_id

        entity_override_map = {
            "Override": "ENTITY_OVERRIDE_MODE_OVERRIDE",
            "Supplement": "ENTITY_OVERRIDE_MODE_SUPPLEMENT",
        }

        st.entity_override_mode = entity_override_map[entity_override_mode]

        entity_list = []

        for k, v in entity_dict.items():
            ent = types.EntityType.Entity()
            ent.value = k
            ent.synonyms = v
            entity_list.append(ent)

        st.entities = entity_list

        return st

    def list_session_entity_types(
        self, session_id: str, environment_id: str = None
    ) -> List[types.SessionEntityType]:
        """Lists all Session Entities currently active in the Session.

        Args:
          session_id, The session to list all session entity types from.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              sessions/<Session ID>``
          environment_id, The Environment associated with the Session ID to
            list all session entity types from. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              environments/<Environment ID>`
            If ``Environment ID`` is not specified, we assume default 'DRAFT'
              environment.
        """
        if environment_id:
            parent_id = self._merge_session_id_and_env_id(
                session_id, environment_id
            )

        else:
            parent_id = session_id

        request = types.session_entity_type.ListSessionEntityTypesRequest()
        request.parent = parent_id

        client_options = self._set_region(session_id)
        client = services.session_entity_types.SessionEntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_session_entity_types(request)

        session_entities = []
        for page in response.pages:
            print(page)
            # for entity in page.entity_types:
            #     session_entities.append(entity)

        return session_entities

    def get_session_entity_type(
        self, session_entity_type_id: str, environment_id: str = None
    ) -> types.SessionEntityType:
        """Retrieves the specified Session Entity Type.

        Args:
          session_entity_type_id, The Session Entity Type ID to retrieve.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
            sessions/<Session ID>/entityTypes/<Entity Type ID>``
          environment_id, The Environment associated with the Session ID to
            list all session entity types from. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              environments/<Environment ID>`
            If ``Environment ID`` is not specified, we assume default 'DRAFT'
              environment.
        """
        if environment_id:
            parent_id = self._merge_session_entity_id_and_env_id(
                session_entity_type_id, environment_id
            )
        else:
            parent_id = session_entity_type_id

        request = types.GetSessionEntityTypeRequest()
        request.name = parent_id

        client_options = self._set_region(session_entity_type_id)
        client = services.session_entity_types.SessionEntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_session_entity_type(request)

        return response

    def create_session_entity_type(
        self, session_id: str, session_entity_type: types.SessionEntityType
    ) -> types.SessionEntityType:
        """Creates a Session Entity Type object from provided inputs.
        Args:
          session_id, The session to list all session entity types from.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              sessions/<Session ID>``
          session_entity_type, The Session Entity Type object to create for the
            provided session. Use `build_session_entity_type` to create a new
            object.

        """
        request = types.CreateSessionEntityTypeRequest()
        request.parent = session_id
        request.session_entity_type = session_entity_type

        client_options = self._set_region(session_id)
        client = services.session_entity_types.SessionEntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.create_session_entity_type(request)

        return response

    def update_session_entity_type(
        self,
        session_entity_type_id: str,
        environment_id: str = None,
        obj: types.SessionEntityType = None,
        **kwargs,
    ) -> types.SessionEntityType:
        """Updates the specified Session Entity Type object.

        This method will update the specific Sesssion Entity Type object based
        on the provided user inputs. If the user provides the entier Session
        Entity Type object, the entire existing object will be updated.
        Alternatively, kwargs can be provided to only update specific portions
        of the Session Entity Type object.

        Ref: https://github.com/googleapis/python-dialogflow-cx/blob/main/
          google/cloud/dialogflowcx_v3beta1/types/session_entity_type.py#L36

        Args:
          session_entity_type_id, the ID of the Session Entity Type to update
          Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              sessions/<Session ID>/entityTypes/<Entity Type ID>`
          environment_id, The Environment associated with the Session Entity
            Type ID to update. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              environments/<Environment ID>`
            If ``Environment ID`` is not specified, we assume default 'DRAFT'
              environment.
          obj, the Session Entity Type object to send as update
        """
        if environment_id:
            parent_id = self._merge_session_entity_id_and_env_id(
                session_entity_type_id, environment_id
            )
        else:
            parent_id = session_entity_type_id

        if obj:
            session_entity_type = obj
            session_entity_type.name = parent_id
        else:
            session_entity_type = self.get_session_entity_type(parent_id)

        # set agent attributes to args
        for key, value in kwargs.items():
            setattr(session_entity_type, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(parent_id)
        client = services.session_entity_types.SessionEntityTypesClient(
            credentials=self.creds, client_options=client_options
        )
        request = types.UpdateSessionEntityTypeRequest()
        request.session_entity_type = parent_id
        request.update_mask = mask

        response = client.update_session_entity_type(request)

        return response

    def delete_session_entity_type(
        self, session_entity_type_id: str, environment_id: str = None
    ) -> str:
        """Deletes the specified Session Entity Type.

        Args:
          session_entity_type_id, the ID of the Session Entity Type to update
            Format:
              ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
                sessions/<Session ID>/entityTypes/<Entity Type ID>`
            environment_id, The Environment associated with the Session Entity
              Type ID to update. Format:
              ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
                environments/<Environment ID>`
              If ``Environment ID`` is not specified, we assume default 'DRAFT'
                environment.
        """
        if environment_id:
            parent_id = self._merge_session_entity_id_and_env_id(
                session_entity_type_id, environment_id
            )

        else:
            parent_id = session_entity_type_id

        client_options = self._set_region(session_entity_type_id)
        client = services.session_entity_types.SessionEntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        request = types.DeleteSessionEntityTypeRequest()
        request.name = parent_id

        client.delete_session_entity_type(request)

        return f"Session Entity Type {session_entity_type_id} successfully "\
            "deleted."



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from google.cloud.dialogflowcx_v3beta1.types import Agent
from google.cloud.dialogflowcx_v3beta1.types import SpeechToTextSettings


class AgentBuilder:
    """Base Class for CX Agent builder."""

    def __init__(self, obj: Agent = None):
        self.proto_obj = None
        if obj:
            self.load_agent(obj)


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_agent_exist()

        logs = self.proto_obj.advanced_settings.logging_settings
        speech_2_txt = self.proto_obj.speech_to_text_settings
        return (
            f"display_name: {self.proto_obj.display_name}"
            f"\ntime_zone: {self.proto_obj.time_zone}"
            f"\ndefault_language_code: {self.proto_obj.default_language_code}"
            f"\ndescription: {self.proto_obj.description}"
            f"\navatar_uri: {self.proto_obj.avatar_uri}"
            "\nenable_speech_adaptation:"
            f" {speech_2_txt.enable_speech_adaptation}"
            "\nenable_spell_correction:"
            f" {self.proto_obj.enable_spell_correction}"
            "\nsupported_language_codes:"
            f" {self.proto_obj.supported_language_codes}"
            "\nenable_stackdriver_logging:"
            f" {logs.enable_stackdriver_logging}"
            "\nenable_interaction_logging:"
            f" {logs.enable_interaction_logging}"
            f"\nsecurity_settings: {self.proto_obj.security_settings}")


    def _check_agent_exist(self):
        """Check if the proto_obj exists otherwise raise an error."""
        if not self.proto_obj:
            raise ValueError(
                "There is no proto_obj!"
                "\nUse create_new_agent or load_agent to continue."
            )
        elif not isinstance(self.proto_obj, Agent):
            raise ValueError(
                "proto_obj is not an Agent type."
                "\nPlease create or load the correct type to continue."
            )


    def load_agent(
        self, obj: Agent, overwrite: bool = False
    ) -> Agent:
        """Load an existing agent to proto_obj for further uses.

        Args:
          obj (Agent):
            An existing Agent obj.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains an Agent.

        Returns:
          An Agent object stored in proto_obj
        """
        if not isinstance(obj, Agent):
            raise ValueError(
                "The object you're trying to load is not an Agent!"
            )
        if self.proto_obj and not overwrite:
            raise Exception(
                "proto_obj already contains an Agent."
                " If you wish to overwrite it, pass overwrite as True."
            )

        if overwrite or not self.proto_obj:
            self.proto_obj = obj

        return self.proto_obj


    def create_new_agent(
        self,
        display_name: str,
        time_zone: str,
        default_language_code: str = "en",
        description: str = None,
        avatar_uri: str = None,
        overwrite: bool = False
    ) -> Agent:
        """Create a new Agent.

        Args:
          display_name (str):
            Required. The human-readable name of the
            agent, unique within the location.
          time_zone (str):
            Required. The time zone of the agent from the
            `time zone database <https://www.iana.org/time-zones>`.
            e.g., America/New_York, Europe/Paris.
          default_language_code (str):
            Required. Immutable. The default language of the agent as a
            language tag. See `Language Support
              <https://cloud.google.com/dialogflow/cx/docs/reference/language>`
            for a list of the currently supported language codes. This
            field cannot be updated.
          description (str):
            The description of the agent. The maximum
            length is 500 characters. If exceeded, the
            request is rejected.
          avatar_uri (str):
            The URI of the agent's avatar. Avatars are used throughout
            the Dialogflow console and in the self-hosted `Web Demo
              <https://cloud.google.com/dialogflow/docs/integrations/web-demo>`
            integration.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains an Agent.

        Returns:
          An Agent object stored in proto_obj
        """
        if ((display_name and not isinstance(display_name, str)) or
            (time_zone and not isinstance(time_zone, str))):
            raise ValueError(
                "display_name and time_zone should be string."
            )
        if self.proto_obj and not overwrite:
            raise Exception(
                "proto_obj already contains an Agent."
                " If you wish to overwrite it, pass overwrite as True."
            )
        if overwrite or not self.proto_obj:
            self.proto_obj = Agent(
                display_name=display_name,
                time_zone=time_zone,
                default_language_code=default_language_code,
                description=description,
                avatar_uri=avatar_uri,
                locked=False,
            )

        return self.proto_obj


    def language_and_speech_settings(
        self,
        enable_speech_adaptation: bool = False,
        enable_spell_correction: bool = False,
        supported_language_codes: List[str] = None,
    ) -> Agent:
        """Change the language and speech settings.

        Args:
          enable_speech_adaptation (bool):
            Whether to use speech adaptation for speech
            recognition.
          enable_spell_correction (bool):
            Indicates if automatic spell correction is
            enabled in detect intent requests.
          supported_language_codes (List[str]):
            The list of all languages supported by the agent
            (except for the ``default_language_code``).

        Returns:
          An Agent object stored in proto_obj
        """
        self._check_agent_exist()

        if isinstance(enable_speech_adaptation, bool):
            self.proto_obj.speech_to_text_settings=SpeechToTextSettings(
                enable_speech_adaptation=enable_speech_adaptation
            )
        if isinstance(enable_spell_correction, bool):
            self.proto_obj.enable_spell_correction = enable_spell_correction
        if supported_language_codes:
            if (isinstance(supported_language_codes, list) and
                all(
                    (isinstance(lang, str) for lang in supported_language_codes)
                )):
                the_obj = self.proto_obj
                the_obj.supported_language_codes = supported_language_codes
            else:
                raise ValueError(
                    "supported_language_codes should be a list of strings."
                )

        return self.proto_obj


    def security_and_logging_settings(
        self,
        enable_stackdriver_logging: bool = False,
        enable_interaction_logging: bool = False,
        security_settings: str = None,
    ) -> Agent:
        """ Change the security and logging settings.

        Args:
          enable_stackdriver_logging (bool):
            If true, StackDriver logging is currently
            enabled.
          enable_interaction_logging (bool):
            If true, DF Interaction logging is currently
            enabled.
          security_settings (str):
            Name of the
            ``[SecuritySettings]
              [google.cloud.dialogflow.cx.v3beta1.SecuritySettings]``
            reference for the agent.Format:
            ``projects/<Project ID>/locations/<Location ID>/
              securitySettings/<Security Settings ID>``.

        Returns:
          An Agent object stored in proto_obj
        """
        self._check_agent_exist()

        logs = self.proto_obj.advanced_settings.logging_settings
        if isinstance(enable_stackdriver_logging, bool):
            logs.enable_stackdriver_logging = enable_stackdriver_logging
        if isinstance(enable_interaction_logging, bool):
            logs.enable_interaction_logging = enable_interaction_logging
        if security_settings and isinstance(security_settings, str):
            self.proto_obj.security_settings = security_settings

        return self.proto_obj


    def show_agent_info(self):
        """Show the proto_obj information."""
        self._check_agent_exist()

        print(self.__str__())



#!/usr/bin/env python
# coding: utf-8

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.intents import Intents
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.pages import Pages
from dfcx_scrapi.tools.dataframe_functions import DataframeFunctions
from dfcx_scrapi.tools.maker_util import MakerUtil


# # Method 1 - Creating an Agent from Simple Text Inputs

# ## Create Your Agent
# Creating an agent requires a minimum of 2 pieces of information:
# - `project_id`, which is your GCP Project ID
# - `display_name`, (i.e. 'My Cool Agent!')
# - `gcp_region`, (Optional) This defaults to `global` region, but you can provide any GCP region that is currently available for Dialogflow CX.

def build_agent(creds_path, project_id, gcp_region, display_name)

	# First we will instantiate our Agent object
	a = Agents(creds_path=creds_path)

	# Next, we will set some variables for our agent creation or retrieval args

	# Then we will call the `create_agent` and capture the result in a var call `my_agent`
	my_agent = a.create_agent(project_id, display_name, gcp_region)

	#Option 2: If agent already exists
	#The agent ID must be entered as a string "projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>"
	#my_agent = a.get_agent("projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>")


	# ## Create Your First Intent
	# For this demo agent, we'll build a basic intent from list of Training Phrases (TPs)
	# 
	# To simplify the Intent creation, we'll utilize the `DataframeFunctions` class from the `tools` portion of the SCRAPI library.   
	# This will allow us to build our intent into a simple Pandas DataFrame, and then push this DataFrame directly into our bot that we just created.
	# 
	# A common method of building Intents and Training Phrases for Dialogflow CX agents is to use Google Sheets or CSVs to store the Intent/TP data.  
	# For this demo, we are working with annotated training phrases. To do so, we've included 
	# * `intent_sample_with_parts.csv` which contains all the training phrases parts. It will be pulled into a dataframe 
	# * `intent_sample_params`, that defines the parameters used for each intent. It will also be pulled into a dataframe

	# First, we will instantiate our DataframeFunctions (dffx) object
	dffx = DataframeFunctions(creds_path=creds_path)

	# Next, we will read in our sample CSV with Intent training phrases data as well as the entities used into 2 distinct Pandas DataFrames
	df = pd.read_csv('../../data/intent_sample_with_parts.csv')
	params_df = pd.read_csv('../../data/intent_sample_params.csv')

	# Finally, we will use `dffx` to push our Intents to our Agent
	#If intents do not exist, use bulk_create_intent_from_dataframe
	my_intents = dffx.bulk_create_intent_from_dataframe(my_agent.name, df, params_df, update_flag=True, mode="advanced")

	#If intent already exist, use bulk_update_intents_from_dataframe
	#my_intents = dffx.bulk_update_intents_from_dataframe(my_agent.name, df, params_df, update_flag=True, mode="advanced")


	# # Bot Building 101 End
	# ## And there you have it!   
	# We've created a simple Dialogflow CX agent using only Python in a Jupyter notebook.
	# You can see how this could be easily scaled up using .py files, git repos, and other scripts to speed up the bot building process.


#The agent ID must be entered as a string "projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>"
def update_agent(creds_path, agent_id)
	# First we will instantiate our Agent object and retrieve the agent using its id
	a = Agents(creds_path=creds_path)
	my_agent = a.get_agent("projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>")


	# ## Create Your First Intent
	# For this demo agent, we'll build a basic intent from list of Training Phrases (TPs)
	# 
	# To simplify the Intent creation, we'll utilize the `DataframeFunctions` class from the `tools` portion of the SCRAPI library.   
	# This will allow us to build our intent into a simple Pandas DataFrame, and then push this DataFrame directly into our bot that we just created.
	# 
	# A common method of building Intents and Training Phrases for Dialogflow CX agents is to use Google Sheets or CSVs to store the Intent/TP data.  
	# For this demo, we are working with annotated training phrases. To do so, we've included 
	# * `intent_sample_with_parts.csv` which contains all the training phrases parts. It will be pulled into a dataframe 
	# * `intent_sample_params`, that defines the parameters used for each intent. It will also be pulled into a dataframe

	# First, we will instantiate our DataframeFunctions (dffx) object
	dffx = DataframeFunctions(creds_path=creds_path)

	# Next, we will read in our sample CSV with Intent training phrases data as well as the entities used into 2 distinct Pandas DataFrames
	df = pd.read_csv('../../data/intent_sample_with_parts.csv')
	params_df = pd.read_csv('../../data/intent_sample_params.csv')

	#bulk update the existing intents
	my_intents = dffx.bulk_update_intents_from_dataframe(my_agent.name, df, params_df, update_flag=True, mode="advanced")


	# # Bot Building 102 End
	# ## And there you have it!   
	# We've created a simple Dialogflow CX agent using only Python in a Jupyter notebook.
	# You can see how this could be easily scaled up using .py files, git repos, and other scripts to speed up the bot building process.

if __name__ == '__main__':
    CREDS_PATH = str(sys.argv[1])
    PROJECT_ID = str(sys.argv[2])
    build_agent(CREDS_PATH, PROJECT_ID)


"""CX Session Resource functions."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types

from dfcx_scrapi.core.scrapi_base import ScrapiBase

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class Sessions(ScrapiBase):
    """Core Class for CX Session Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        scope=False,
        agent_id: str = None,
        session_id: str = None
    ):
        super().__init__(
            creds_path=creds_path, creds_dict=creds_dict, scope=scope
        )

        self.session_id = session_id
        self.agent_id = agent_id

    @property
    def session_id(self):
        return self._session_id

    @session_id.setter
    def session_id(self, value):
        if value:
            self._parse_resource_path("session", value)

        self._session_id = value

    @staticmethod
    def _build_query_input(text, language_code):
        """Build out the query_input object for the Query Request.

        Args:
          text, the text to use for the Detect Intent request.
          language_code, the language code to use for Detect Intent request.
          """
        text_input = types.session.TextInput(text=text)
        query_input = types.session.QueryInput(
            text=text_input, language_code=language_code)

        return query_input

    def build_session_id(
        self, agent_id:str = None, overwrite:bool = True) -> str:
        """Creates a valid UUID-4 Session ID to use with other methods.

        Args:
          overwrite (Optional), if a session_id already exists, this will
            overwrite the existing Session ID parameter. Defaults to True.
        """

        agent_parts = self._parse_resource_path("agent", agent_id)
        session_id = f"projects/{agent_parts['project']}/"\
            f"locations/{agent_parts['location']}/agents/"\
            f"{agent_parts['agent']}/sessions/{uuid.uuid4()}"

        if overwrite:
            self.session_id = session_id

        return session_id

    def run_conversation(
        self,
        agent_id: str = None,
        session_id: str = None,
        conversation: List[str] = None,
        parameters=None,
        response_text=False,
    ):
        """Tests a full conversation with the specified CX Agent.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          conversation: a List of Strings that represent the USER utterances
            for the given conversation, in the order they would happen
            chronologically in the conversation.
            Ex:
              ['I want to check my bill', 'yes', 'no that is all', 'thanks!']
          parameters: (Optional) Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.
          response_text: Will provide the Agent Response text if set to True.
            Default value is False.

        Returns:
          None, the conversation Request/Response is printed to console.
        """
        if not session_id:
            session_id = self.session_id

        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )
        session_path = f"{agent_id}/sessions/{session_id}"

        if parameters:
            query_params = types.session.QueryParameters(parameters=parameters)
            text_input = types.session.TextInput(text="")
            query_input = types.session.QueryInput(
                text=text_input, language_code="en"
            )
            request = types.session.DetectIntentRequest(
                session=session_path,
                query_params=query_params,
                query_input=query_input,
            )

            response = session_client.detect_intent(request=request)

        for text in conversation:
            text_input = types.session.TextInput(text=text)
            query_input = types.session.QueryInput(
                text=text_input, language_code="en"
            )
            request = types.session.DetectIntentRequest(
                session=session_path, query_input=query_input
            )
            response = session_client.detect_intent(request=request)
            query_result = response.query_result

            print("=" * 20)
            print(f"Query text: {query_result.text}")
            if "intent" in query_result:
                print(
                    f"Triggered Intent: {query_result.intent.display_name}"
                )

            if "intent_detection_confidence" in query_result:
                print(
                    f"Intent Confidence: \
                        f{query_result.intent_detection_confidence}"
                )

            print(
                f"Response Page: {query_result.current_page.display_name}"
            )

            for param in query_result.parameters:
                if param == "statusMessage":
                    print(
                        f"Status Message: {query_result.parameters[param]}"
                    )

            if response_text:
                concat_messages = " ".join(
                    [
                        " ".join(response_message.text.text)
                        for response_message in query_result.response_messages
                            ]
                        )
                print(f"Response Text: {concat_messages}\n")


    def detect_intent(
        self,
        agent_id,
        session_id,
        text,
        language_code="en",
        parameters=None):
        """Returns the result of detect intent with texts as inputs.

        Using the same `session_id` between requests allows continuation
        of the conversation.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          text: the user utterance to run intent detection on
          parameters: (Optional) Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.

        Returns:
          The CX query result from intent detection
        """
        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )

        res = self._parse_resource_path("session", str(session_id), False)
        if not res:
            raise ValueError(
                "Session ID must be provided in the following format: "
                "`projects/<Project ID>/locations/<Location ID>/agents/"\
                "<Agent ID>/sessions/<Session ID>`.\n\n"\
                "Utilize `build_session_id` to create a new Session ID.")

        logging.info(f"Starting Session ID {session_id}")

        if parameters:
            query_params = types.session.QueryParameters(parameters=parameters)

            query_input = self._build_query_input(text, language_code)

            request = types.session.DetectIntentRequest()
            request.session = session_id
            request.query_input = query_input
            request.query_params = query_params

            response = session_client.detect_intent(request=request)

        else:
            query_input = self._build_query_input(text, language_code)

            request = types.session.DetectIntentRequest()
            request.session = session_id
            request.query_input = query_input

        response = session_client.detect_intent(request)
        query_result = response.query_result

        return query_result

    def preset_parameters(
        self, agent_id: str = None, session_id: str = None, parameters=None
    ):
        """Used to set session parameters before a conversation starts.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          parameters: Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.

        Returns:
          The CX query result from intent detection run with no text input
        """
        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )
        session_path = f"{agent_id}/sessions/{session_id}"

        query_params = types.session.QueryParameters(parameters=parameters)
        text_input = types.session.TextInput(text=None)
        query_input = types.session.QueryInput(
            text=text_input, language_code="en"
        )
        request = types.session.DetectIntentRequest(
            session=session_path,
            query_params=query_params,
            query_input=query_input,
        )

        response = session_client.detect_intent(request=request)

        return response



"""CX Versions Resource functions."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict
from google.oauth2 import service_account
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Versions(scrapi_base.ScrapiBase):
    """Core Class for CX Versions Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict[str,str] = None,
        creds: service_account.Credentials = None,
        flow_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
        )

        if flow_id:
            self.flow_id = flow_id

    def list_versions(self, flow_id:str):
        """List all Versions for a given Flow.

        Args:
          flow_id: Required. The targeted flow for the operation. Format:
            projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>
              /flows/<Flow ID>

        Returns:
          List of Version objects.
        """
        if not flow_id:
            flow_id = self.flow_id

        request = types.version.ListVersionsRequest()
        request.parent = flow_id

        client_options = self._set_region(flow_id)
        client = services.versions.VersionsClient(
            client_options=client_options, credentials=self.creds
        )

        response = client.list_versions(request)

        versions = []
        for page in response.pages:
            for version in page.versions:
                versions.append(version)

        return versions

    def get_version(
        self,
        version_id:str=None,
        display_name:str=None,
        flow_id:str=None):
        """Get Version object for specific version.

        Requires either Version's ID or Version's display name.
        If both are provided, display_name is considered first.

        Args:
          version_id: Unique Version ID of the target to get. Format:
            projects/<Project ID>/locations/<Location ID>/agents/
            <Agent ID>/flows/<Flow ID>/versions/<Version ID>
          display_name: Human readable display name of the Version to get.
          flow_id: The targeted flow for the operation. format:
          projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
            flows/<Flow ID>

        Returns:
          Version object.
        """

        if not display_name and not version_id:
            logging.warning("versions.get_version requires version_id or"
            " display_name.")
            return None

        if display_name:
            response = self.get_version_by_display_name(display_name, flow_id)

        else:
            request = types.version.GetVersionRequest(name=version_id)
            client = services.versions.VersionsClient(
                client_options=self._set_region(version_id),
                credentials=self.creds
            )

            response = client.get_version(request)

        return response

    def get_version_by_display_name(self, display_name:str, flow_id:str):
        """Get Version object for specific version by its display name.

        Args:
          display_name: Human readable display name of the target to get.
          flow_id: The targeted flow for the operation. format:
          projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
            flows/<Flow ID>

        Returns:
          Version object.
        """
        versions_list = self.list_versions(flow_id)
        for version_obj in versions_list:
            if version_obj.display_name == display_name:
                return version_obj

        return None

    def load_version(
        self,
        version:types.version.Version,
        allow_override:bool = False,
        flow_id:str = None):
        """Switch a flow to the specified version.

        Args:
          version: Version object of the desired target version.
          allow_override: allow_override_agent_resources is false,
                conflicted agent-level resources will not be overridden
                (i.e. intents, entities, webhooks)
          flow_id: The targeted flow for the operation. Format:
            projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>

        Returns:
          An object representing a long-running operation (LRO).
        """

        if not flow_id:
            flow_id = self.flow_id

        request = types.version.LoadVersionRequest(
            name=version.name,
            allow_override_agent_resources=allow_override
        )

        client_options = self._set_region(flow_id)
        client = services.versions.VersionsClient(
            client_options=client_options, credentials=self.creds
        )

        response = client.load_version(request)
        return response

    def create_version(
        self,
        flow_id:str,
        display_name:str,
        description:str=None):
        """Create a Version for the specified Flow ID.

        Args:
          flow_id: The targeted flow for the operation. Format:
            projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>
          display_name: Human readable display name of the Version created.
          description: Additional description details for the Version created.

        Returns:
          An object representing a long-running operation (LRO).
        """

        request = types.version.CreateVersionRequest()

        client_options = self._set_region(flow_id)
        client = services.versions.VersionsClient(
            client_options=client_options, credentials=self.creds
        )

        version = types.Version()
        version.display_name = display_name
        version.description = description

        request.parent = flow_id
        request.version = version

        response = client.create_version(request)

        return response

    def delete_version(self, version_id:str):
        """Delete a specified Version.

        Args:
          version_name: Unique Version ID of the target to delete. Format:
            projects/<Project ID>/locations/<Location ID>/agents/
            <Agent ID>/flows/<Flow ID>/versions/<Version ID>
        """
        request = types.version.DeleteVersionRequest(name=version_id)

        client = services.versions.VersionsClient(
            client_options=self._set_region(version_id),
            credentials=self.creds
        )

        return client.delete_version(request)

    def compare_versions(
        self,
        base_version_id:str,
        target_version_id:str,
        flow_id:str = None):
        """Compares the specified base version with target version.

        Args:
          base_version_id: ID of the base flow version
            to compare with the target version. format:
            projects/<Project ID>/locations/<Location ID>/agents/
            <Agent ID>/flows/<Flow ID>/versions/<Version ID>.
          target_version_id: ID of the target flow version
            to compare with the base version. format:
            projects/<Project ID>/locations/<Location ID>/agents/
            <Agent ID>/flows/<Flow ID>/versions/<Version ID>.
          flow_id: Required. The targeted flow for the operation. Format:
            projects/<Project ID>/locations/<Location ID>/
            agents/<Agent ID>/flows/<Flow ID>

        Returns:
          types.CompareVersionsRequest Object.
        """
        if not flow_id:
            flow_id = self.flow_id

        request = types.version.CompareVersionsRequest(
            base_version=base_version_id,
            target_version=target_version_id
        )

        client_options = self._set_region(flow_id)
        client = services.versions.VersionsClient(
            client_options=client_options, credentials=self.creds
        )

        response = client.compare_versions(request)

        return response



"""Intent Resource functions."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import json
import logging
from typing import Dict, List, Tuple
import pandas as pd

from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core.scrapi_base import ScrapiBase

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Intents(ScrapiBase):
    """Core Class for CX Intent Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        intent_id: str = None,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if intent_id:
            self.intent_id = intent_id
            self.client_options = self._set_region(self.intent_id)

        if agent_id:
            self.agent_id = agent_id


    @staticmethod
    def intent_proto_to_dataframe(
        obj: types.Intent, mode: str = "basic"
    ) -> pd.DataFrame:
        """Converts an intent protobuf object to a Pandas DataFrame.

        Args:
          obj (types.Intent): the intent protobuf object
          mode (str):
            "basic" returns display name and training phrase as plain text.
            "advanced" returns training phrases broken out by parts
            with their parameters included.

        Returns:
          In basic mode, a Pandas DataFrame with columns:
            display_name, training phrase
          In advanced mode, a Pandas DataFrame with columns:
            name, display_name, description, priority,
            is_fallback, labels, id, repeat_count,
            training_phrase_idx, text, text_idx,
            parameter_id, entity_type, is_list, redact
        """
        if not isinstance(obj, types.Intent):
            raise ValueError("obj should be Intent.")

        if mode == "basic":
            df = pd.DataFrame(columns=["display_name", "training_phrase"])

            intent_dict = {"display_name": str(obj.display_name)}

            if not obj.training_phrases:
                row = pd.DataFrame.from_dict(
                    intent_dict, orient="index"
                ).transpose()
                df = pd.concat([df, row], ignore_index=True)
            else:
                for tp in obj.training_phrases:
                    parts_list = [part.text for part in tp.parts]
                    intent_dict.update({"training_phrase": "".join(parts_list)})

                    row = pd.DataFrame.from_dict(
                        intent_dict, orient="index"
                    ).transpose()
                    df = pd.concat([df, row], ignore_index=True)

            return df

        elif mode == "advanced":
            df = pd.DataFrame(columns=[
                "name", "display_name", "description", "priority",
                "is_fallback", "labels", "id", "repeat_count",
                "training_phrase_idx", "text", "text_idx",
                "parameter_id", "entity_type", "is_list", "redact",
            ])

            intent_dict = {
                "name": str(obj.name),
                "display_name": str(obj.display_name),
                "description": str(obj.description),
                "priority": int(obj.priority),
                "is_fallback": bool(obj.is_fallback),
            }

            # labels
            intent_dict["labels"] = ",".join([
                key if key == val else f"{key}:{val}"
                for key, val in obj.labels.items()
            ])
            # parameters
            params_dict = {
                str(param.id): {
                    "parameter_id": str(param.id),
                    "entity_type": str(param.entity_type),
                    "is_list": bool(param.is_list),
                    "redact": bool(param.redact),
                }
                for param in obj.parameters
            }
            # training phrases
            if not obj.training_phrases:
                row = pd.DataFrame.from_dict(
                    intent_dict, orient="index"
                ).transpose()
                df = pd.concat([df, row], ignore_index=True)
            else:
                for tp_count, tp in enumerate(obj.training_phrases):
                    intent_dict.update({
                        "id": str(tp.id),
                        "repeat_count": int(tp.repeat_count),
                        "training_phrase_idx": tp_count,
                    })
                    for part_count, part in enumerate(tp.parts):
                        intent_dict.update({
                            "text": part.text,
                            "text_idx": part_count
                        })
                        if part.parameter_id:
                            intent_dict.update(params_dict[part.parameter_id])
                        elif intent_dict.get("entity_type"):
                            # Remove existing parameter_id if exist
                            key_to_remove = [
                                "parameter_id", "entity_type",
                                "is_list", "redact",
                            ]
                            for key in key_to_remove:
                                intent_dict.pop(key)
                        # Add to Dataframe
                        row = pd.DataFrame.from_dict(
                            intent_dict, orient="index"
                        ).transpose()
                        df = pd.concat([df, row], ignore_index=True)
            return df
        else:
            raise ValueError("Mode types: [basic, advanced]")


    @staticmethod
    def modify_training_phrase_df(
        actions: pd.DataFrame, training_phrase_df: pd.DataFrame
    ):
        """Update the advanced mode training phrases dataframe to reflect the
        actions provided in the actions dataframe. Pass the returned new
        training phrase dataframe and the original parameters dataframe to the
        bulk_update_intents_from_dataframe functions with update_flag = True in
        order to make these edits. Entities in training phrases not touched
        will be maintained. Entities will not be added in phrases which are
        added or moved.

        Args:
          actions: A Pandas DataFrame with columns:
            display_name: the display name of the intent to take action on
            phrase: exact string training phrase to take action on
            action: "add" or "delete". To do a move, it will be an add and
              a delete action of the same phrase
          training_phrase_df: advanced mode training phrase dataframe pulled

        Returns:
          A dictionary with keys updated_training_phrases_df and actions_taken.
            The value for updated_training_phrases_df is the training phrase
              DataFrame from advanced mode with the edits made to it shown
              in the actions_taken DataFrame.
            The value for actions_taken is a DataFrame of actions taken based
              on the actions provided. For example, we cannot delete a training
              phrase which does not exist; this will be shown.
        """
        mutations = pd.merge(
            training_phrase_df,
            actions,
            on=["display_name", "phrase"],
            how="outer",
        )

        # untouched
        untouched = mutations[mutations["action"].isna()]

        # Adding new phrases
        true_additions = mutations[
            (mutations["action"] == "add") & (mutations["text"].isna())
        ]
        false_additions = mutations[
            (mutations["action"] == "add") & (~mutations["text"].isna())
        ]

        true_additions = true_additions.drop(
            columns=[
                "name",
                "training_phrase",
                "part",
                "text",
                "parameter_id",
                "repeat_count",
                "id",
            ]
        )
        true_additions["name"] = list(set(untouched["name"]))[0]

        next_tp_id = int(untouched["training_phrase"].max() + 1)
        last_tp_id = int(
            untouched["training_phrase"].max() + len(true_additions)
        )

        true_additions["training_phrase"] = list(
            range(next_tp_id, last_tp_id + 1)
        )
        true_additions["part"] = 0
        true_additions["text"] = true_additions["phrase"]
        true_additions["parameter_id"] = ""
        true_additions["repeat_count"] = 1
        true_additions["id"] = ""
        true_additions = true_additions[untouched.columns]

        # Deleting existing phrases
        true_deletions = mutations.copy()[
            (mutations["action"] == "delete") & (~mutations["text"].isna())
        ]
        false_deletions = mutations.copy()[
            (mutations["action"] == "delete") & (mutations["text"].isna())
        ]

        updated_training_phrases_df = untouched.copy()
        updated_training_phrases_df = pd.concat([
            updated_training_phrases_df,
            false_additions
        ])
        updated_training_phrases_df = pd.concat([
            updated_training_phrases_df,
            true_additions
        ])
        updated_training_phrases_df = updated_training_phrases_df.drop(
            columns=["action"]
        )

        actions_taken = pd.DataFrame()
        if true_additions.empty is False:
            true_additions.insert(len(true_additions.columns), "completed", 1)
            true_additions.insert(
                len(true_additions.columns), "category", "true addition"
            )
            true_additions.insert(
                len(true_additions.columns),
                "outcome",
                true_additions.apply(
                    lambda x: "{} added to {}".format( #pylint: disable=C0209
                        x["phrase"], x["display_name"]
                    ),
                    axis=1,
                ),
            )
            actions_taken = pd.concat([actions_taken, true_additions])

        if false_additions.empty is False:
            false_additions.insert(len(false_additions.columns), "completed", 0)
            false_additions.insert(
                len(false_additions.columns), "category", "false addition"
            )
            false_additions.insert(
                len(false_additions.columns),
                "outcome",
                false_additions.apply(
                    lambda x: "{} already in {}".format( #pylint: disable=C0209
                        x["phrase"], x["display_name"]
                    ),
                    axis=1,
                ),
            )
            actions_taken = pd.concat([actions_taken, false_additions])

        if true_deletions.empty is False:
            true_deletions.insert(len(true_deletions.columns), "completed", 1)
            true_deletions.insert(
                len(true_deletions.columns), "category", "true deletion"
            )
            true_deletions.insert(
                len(true_deletions.columns),
                "outcome",
                true_deletions.apply(
                    lambda x: "{} removed from {}".format( #pylint: disable=C0209
                        x["phrase"], x["display_name"]
                    ),
                    axis=1,
                ),
            )
            actions_taken = pd.concat([actions_taken, true_deletions])

        if false_deletions.empty is False:
            false_deletions.insert(len(false_deletions.columns), "completed", 0)
            false_deletions.insert(
                len(false_deletions.columns), "category", "false deletion"
            )
            false_deletions.insert(
                len(false_deletions.columns),
                "outcome",
                false_deletions.apply(
                    lambda x: "{} not found in {}".format( #pylint: disable=C0209
                        x["phrase"], x["display_name"]
                    ),
                    axis=1,
                ),
            )
            actions_taken = pd.concat([actions_taken, false_deletions])

        actionable_intents = list(
            set(actions_taken[actions_taken["completed"] == 1]["display_name"])
        )

        updated_training_phrases_df = updated_training_phrases_df[
            updated_training_phrases_df["display_name"].isin(actionable_intents)
        ]
        return_data = {
            "updated_training_phrases_df": updated_training_phrases_df,
            "actions_taken": actions_taken,
        }

        return return_data

    def get_intents_map(self, agent_id: str = None, reverse: bool = False):
        """Exports Agent Intent Names and UUIDs into a user friendly dict.

        Args:
          agent_id: the formatted CX Agent ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Intent UUIDs as keys and display names as values
        """
        if not agent_id:
            agent_id = self.agent_id

        if reverse:
            intents_dict = {
                intent.display_name: intent.name
                for intent in self.list_intents(agent_id)
            }

        else:
            intents_dict = {
                intent.name: intent.display_name
                for intent in self.list_intents(agent_id)
            }

        return intents_dict

    def list_intents(
        self,
        agent_id: str = None,
        language_code: str = None) -> List[types.Intent]:
        """Exports List of all intents in specific CX Agent.

        Args:
          agent_id: the formatted CX Agent ID to use
          language_code: Language code of the intents being uploaded. Ref:
            https://cloud.google.com/dialogflow/cx/docs/reference/language

        Returns:
          List of Intent objects
        """
        if not agent_id:
            agent_id = self.agent_id

        request = types.intent.ListIntentsRequest()

        if language_code:
            request.language_code = language_code

        request.parent = agent_id
        client_options = self._set_region(agent_id)
        client = services.intents.IntentsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_intents(request)

        intents = []
        for page in response.pages:
            for intent in page.intents:
                intents.append(intent)

        return intents

    def get_intent(
        self,
        intent_id: str = None,
        language_code: str = None) -> types.Intent:
        """Get a single Intent object based on specific CX Intent ID.

        Args:
          intent_id: the properly formatted CX Intent ID
          language_code: Language code of the intents being uploaded. Ref:
            https://cloud.google.com/dialogflow/cx/docs/reference/language

        Returns:
          A single Intent object
        """
        if not intent_id:
            intent_id = self.intent_id

        request = types.intent.GetIntentRequest()

        if language_code:
            request.language_code = language_code

        request.name = intent_id
        client_options = self._set_region(intent_id)
        client = services.intents.IntentsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_intent(request)

        return response

    def create_intent(
        self,
        agent_id: str,
        obj: types.Intent = None,
        intent_dictionary: dict = None,
        language_code: str = None) -> types.Intent:
        """Creates an Intent from a protobuf or dictionary.

        Args:
          agent_id: the formatted CX Agent ID to use
          obj: (Optional) Intent protobuf of types.Intent
          intent_dictionary: (optional) dictionary of the intent to pass in
            with structure

        example intent dictionary:
           test_intent = {
            "description": "",
            "display_name": "my_intent",
            "is_fallback": False,
            "labels": {},
            "priority": 500000,
            "training_phrases": [
                {
                    "id": "",
                    "parts": [
                        {
                            "text": "hello"
                        },
                        {
                            "text": "all"
                        }
                    ],
                    "repeat_count": 1
                },
                {
                    "id": "",
                    "parts": [
                        {
                            "text": "hi"
                        }
                    ],
                    "repeat_count": 1
                }
            ]
        }

        Returns:
          Intent protobuf object
        """

        if obj and intent_dictionary:
            raise ValueError("cannot provide both obj and intent_dictionary")
        elif obj:
            intent = obj
            intent.name = ""
        elif intent_dictionary:
            intent = types.intent.Intent.from_json(
                json.dumps(intent_dictionary)
            )
        else:
            raise ValueError("must provide either obj or intent_dictionary")

        request = types.intent.CreateIntentRequest()

        if language_code:
            request.language_code = language_code

        request.parent = agent_id
        request.intent = intent

        client_options = self._set_region(agent_id)
        client = services.intents.IntentsClient(
            client_options=client_options, credentials=self.creds
        )

        response = client.create_intent(request)

        return response

    def update_intent(
        self,
        intent_id: str = None,
        obj: types.Intent = None,
        language_code: str = None,
        **kwargs) -> types.Intent:
        """Updates a single Intent object based on provided arguments.

        Args:
          intent_id: the destination Intent ID. Must be formatted properly
            for Intent IDs in CX.
          obj: The CX Intent object in proper format. This can also be
            extracted by using the get_intent() method.
          language_code: Language code of the intents being uploaded. Ref:
            https://cloud.google.com/dialogflow/cx/docs/reference/language

        Returns:
          The updated intent object.
        """
        if obj:
            intent = obj
            intent.name = intent_id

        else:
            if not intent_id:
                intent_id = self.intent_id
            intent = self.get_intent(intent_id)

        # set intent attributes from kwargs
        for key, value in kwargs.items():
            setattr(intent, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(intent_id)
        client = services.intents.IntentsClient(
            client_options=client_options, credentials=self.creds
        )

        request = types.intent.UpdateIntentRequest()

        request.intent = intent
        request.update_mask = mask

        if language_code:
            request.language_code = language_code

        response = client.update_intent(request)

        return response

    def delete_intent(self, intent_id: str, obj: types.Intent = None) -> None:
        """Deletes an intent by Intent ID.

        Args:
          intent_id: intent to delete
        """
        if obj:
            intent_id = obj.name
        else:
            client_options = self._set_region(intent_id)
            client = services.intents.IntentsClient(
                client_options=client_options, credentials=self.creds
            )
            client.delete_intent(name=intent_id)

    def bulk_intent_to_df(
        self,
        agent_id: str = None,
        mode: str = "basic",
        intent_subset: List[str] = None,
        transpose: bool = False,
        language_code: str = None) -> pd.DataFrame:
        """Extracts all Intents and Training Phrases into a Pandas DataFrame.

        Args:
          agent_id (str):
            agent to pull list of intents
          mode (str):
            "basic" returns display name and training phrase as plain text.
            "advanced" returns training phrases broken out by parts
            with their parameters included.
          intent_subset (List[str]):
            A subset of intents to extract the intents from.
          transpose (bool):
            Return the transposed DataFrame. If this flag passed as True,
            mode won't affect the result and the result would be like basic.
          language_code (str):
            Language code of the intents being uploaded. Ref:
            https://cloud.google.com/dialogflow/cx/docs/reference/language

        Returns:
          In basic mode, a Pandas DataFrame with columns:
            display_name, training phrase
          In advanced mode, a Pandas DataFrame with columns:
            name, display_name, description, priority,
            is_fallback, labels, id, repeat_count,
            training_phrase_idx, text, text_idx,
            parameter_id, entity_type, is_list, redact
        """

        if not agent_id:
            agent_id = self.agent_id

        if transpose:
            _, intents_dict = self.intents_to_df_cosine_prep(agent_id)
            transposed_df = pd.DataFrame.from_dict(
                intents_dict, "index"
            ).transpose()
            if intent_subset:
                transposed_df = transposed_df[intent_subset]

            return transposed_df

        if mode not in ["basic", "advanced"]:
            raise ValueError("Mode types: [basic, advanced]")

        main_df = pd.DataFrame()
        intents = self.list_intents(agent_id, language_code=language_code)

        for obj in intents:
            if (intent_subset) and (obj.display_name not in intent_subset):
                continue
            intent_df = self.intent_proto_to_dataframe(obj, mode=mode)
            main_df = pd.concat([main_df, intent_df], ignore_index=True)

        return main_df

    def intents_to_df_cosine_prep(
        self,
        agent_id: str = None) -> Tuple[pd.DataFrame, Dict[str,str]]:
        """Exports a dataframe and defaultdict of Intents for use with Cosine
        Similarity tools.

        Args:
          agent_id: agent to pull list of intents from

        Returns:
          A tuple containing two values:
            A Pandas Dataframe of Intents and TPs
            A Defaultdict(List) that is prepped to feed to the
              Cosine similarity tool (offline)
        """
        if not agent_id:
            agent_id = self.agent_id

        intent_dict = defaultdict(list)
        intents = self.list_intents(agent_id)

        for intent in intents:  # pylint: disable=R1702
            if intent.display_name == "Default Negative Intent":
                pass
            else:
                if "training_phrases" in intent:
                    for training_phrase in intent.training_phrases:
                        text_list = []
                        if len(training_phrase.parts) > 1:
                            for item in training_phrase.parts:
                                text_list.append(item.text)
                            intent_dict[intent.display_name].append(
                                "".join(text_list))
                        else:
                            intent_dict[intent.display_name].append(
                                training_phrase.parts[0].text
                            )
                else:
                    intent_dict[intent.display_name].append("")

        dataframe = pd.DataFrame.from_dict(
            intent_dict, orient="index").transpose()
        dataframe = dataframe.stack().to_frame().reset_index(level=1)
        dataframe = dataframe.rename(
            columns={"level_1": "display_name",
            0: "training_phrase"}).reset_index(drop=True)
        dataframe = dataframe.sort_values(["display_name", "training_phrase"])

        return dataframe, intent_dict



"""A set of Utility methods to copy DFCX Resources."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import time
from typing import Dict, List
from collections import defaultdict

from google.cloud.dialogflowcx_v3beta1 import types
from google.api_core import exceptions as core_exceptions

from dfcx_scrapi.core.scrapi_base import ScrapiBase
from dfcx_scrapi.core.intents import Intents
from dfcx_scrapi.core.entity_types import EntityTypes
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.pages import Pages
from dfcx_scrapi.core.webhooks import Webhooks
from dfcx_scrapi.core.transition_route_groups import TransitionRouteGroups

# logging config
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)-8s %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)

class CopyUtil(ScrapiBase):
    """Utility class for copying DFCX Resources between Agents."""
    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds = None,
        scope = False,
        agent_id: str = None
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id

        self.intents = Intents(creds=self.creds, agent_id=self.agent_id)
        self.entities = EntityTypes(creds=self.creds, agent_id=self.agent_id)
        self.flows = Flows(creds=self.creds, agent_id=self.agent_id)
        self.pages = Pages(creds=self.creds)
        self.webhooks = Webhooks(creds=self.creds, agent_id=self.agent_id)
        self.route_groups = TransitionRouteGroups(
            creds=self.creds, agent_id=self.agent_id)

    @staticmethod
    def _get_entry_webhooks(page_object, resources):
        """Check the Entry Fulfillment for webhooks and return them."""
        if 'entry_fulfillment' in page_object:
            if 'webhook' in page_object.entry_fulfillment:
                resources['webhooks'].append(
                    page_object.entry_fulfillment.webhook)

        return resources


    @staticmethod
    def _get_condition_route_webhooks(page_object, resources):
        """Extract Webhooks from Condition for a given Page."""
        if 'transition_routes' in page_object:
            for transition_route in page_object.transition_routes:
                if ('condition' in transition_route
                and 'webhook' in transition_route.trigger_fulfillment):
                    resources['webhooks'].append(
                        transition_route.trigger_fulfillment.webhook
                    )

        return resources

    @staticmethod
    def _get_form_entity_types(page_object, resources):
        """Extract Entity Types from Parameters for a given Page."""
        if 'form' in page_object:
            if 'parameters' in page_object.form:
                for param in page_object.form.parameters:
                    if 'sys.' in param.entity_type:
                        continue
                    resources['entities'].append(
                            param.entity_type
                        )

        return resources

    @staticmethod
    def _get_intent_route_intents(page_object, resources):
        """Extract any Intents from Transition Routes in a given Page."""

        if 'transition_routes' in page_object:
            for transition_route in page_object.transition_routes:
                if 'intent' in transition_route:
                    resources['intents'].append(
                        transition_route.intent
                    )

        return resources

    @staticmethod
    def _convert_entry_webhooks(page_object, webhooks_map):
        """Convert webhooks in the entry fulfillment of the Page Object.

        Internal method to convert webhooks in the entry fulfillment of
        a given page from the Resource ID to their Display Name or vice versa.
        """
        if 'webhook' in page_object.entry_fulfillment:
            page_object.entry_fulfillment.webhook = webhooks_map[
                page_object.entry_fulfillment.webhook
            ]

        return page_object

    @staticmethod
    def __convert_tr_target_page(
        trans_route,
        pages_map,
        convert_type = None,
        flows_map = None,
        flow = None):

        if convert_type == 'source':
            if trans_route.target_page.split('/')[-1] == 'END_FLOW':
                trans_route.target_page = 'END_FLOW'

            elif trans_route.target_page.split('/')[-1] == 'END_SESSION':
                trans_route.target_page = 'END_SESSION'

            elif (
                trans_route.target_page.split('/')[-1] == 'CURRENT_PAGE'
            ):
                trans_route.target_page = 'CURRENT_PAGE'

            else:
                trans_route.target_page = pages_map[trans_route.target_page]

        elif convert_type == 'destination':
            if trans_route.target_page == 'END_FLOW':
                trans_route.target_page = (
                    flows_map[flow] + '/pages/END_FLOW')

            elif trans_route.target_page == 'END_SESSION':
                trans_route.target_page = (
                    flows_map[flow] + '/pages/END_SESSION')

            elif trans_route.target_page == 'CURRENT_PAGE':
                trans_route.target_page = (
                    flows_map[flow] + '/pages/CURRENT_PAGE')

            else:
                trans_route.target_page = pages_map[trans_route.target_page]

        return trans_route.target_page

    @staticmethod
    def __convert_trigger_fulfillment_webhook(handler, webhooks_map):
        if 'webhook' in handler.trigger_fulfillment:
            handler.trigger_fulfillment.webhook = webhooks_map[
                handler.trigger_fulfillment.webhook]

        return handler.trigger_fulfillment.webhook

    def _convert_form_parameters( #pylint: disable=too-many-arguments
        self,
        page_object,
        pages_map,
        webhooks_map,
        entities_map,
        convert_type = None):

        for param in page_object.form.parameters:
            if 'fill_behavior' in param:
                if 'initial_prompt_fulfillment' in param.fill_behavior:
                    if 'webhook' in (
                        param.fill_behavior.initial_prompt_fulfillment):
                        param.fill_behavior.initial_prompt_fulfillment.webhook = webhooks_map[ #pylint: disable=line-too-long
                            param.fill_behavior.initial_prompt_fulfillment.webhook #pylint: disable=line-too-long
                        ]

                if 'reprompt_event_handlers' in param.fill_behavior:
                    for handler in param.fill_behavior.reprompt_event_handlers:
                        if 'trigger_fulfillment' in handler:
                            handler.trigger_fulfillment.webhook = (
                                self.__convert_trigger_fulfillment_webhook(
                                    handler, webhooks_map)
                            )

                        if 'target_page' in handler:
                            handler.target_page = (
                                self.__convert_tr_target_page(
                                    handler, pages_map,
                                    convert_type=convert_type)
                            )

            if 'sys.' in param.entity_type:
                pass
            else:
                param.entity_type = entities_map[param.entity_type]

        return page_object

    def _convert_event_handlers(
        self, page_object, pages_map, webhooks_map, convert_type = None):

        for handler in page_object.event_handlers:
            if 'target_page' in handler:
                handler.target_page = self.__convert_tr_target_page(
                    handler, pages_map, convert_type=convert_type)

            if 'trigger_fulfillment' in handler:
                handler.trigger_fulfillment.webhook = (
                    self.__convert_trigger_fulfillment_webhook(
                        handler, webhooks_map
                    )
                )

        return page_object

    def _convert_trans_routes( #pylint: disable=too-many-arguments
        self,
        page_object,
        pages_map,
        intents_map,
        webhooks_map,
        convert_type = None):

        for trans_route in page_object.transition_routes:
            if 'target_page' in trans_route:
                trans_route.target_page = self.__convert_tr_target_page(
                    trans_route, pages_map, convert_type=convert_type)

            if 'intent' in trans_route:
                trans_route.intent = intents_map[trans_route.intent]
                if 'webhook' in trans_route.trigger_fulfillment:
                    trans_route.trigger_fulfillment.webhook = webhooks_map[
                        trans_route.trigger_fulfillment.webhook
                    ]
            elif (
                'condition' in trans_route
                and 'webhook' in trans_route.trigger_fulfillment
            ):
                trans_route.trigger_fulfillment.webhook = webhooks_map[
                    trans_route.trigger_fulfillment.webhook
                ]

        return page_object

    def _get_intent_entity_dependencies(self, resources):
        """Loop through Intents and find any additional Entity dependencies"""
        agent = '/'.join(resources['intents'][0].split('/')[0:6])
        temp_intents = self.intents.list_intents(agent)

        for intent in temp_intents:
            if intent.name in resources['intents']:
                if len(intent.parameters) > 0:
                    for param in intent.parameters:
                        if 'sys.' in param.entity_type:
                            continue
                        resources['entities'].append(param.entity_type)

        return resources

    def _get_flow_intent_route_intents(self, flow_id, obj_list, resources):
        """Extract Intents from Transition Routes on the Flow Start Page.

        The Start Page of every Flow is a special Page-like object that
        actually exists as part of the Flow object. Because it is different
        enough from a standard Page object, we have to extract the resources
        from it differently.
        """
        source_flow = self.flows.get_flow(flow_id)
        temp_page_name_list = [page.name for page in obj_list]

        for transition_route in source_flow.transition_routes:
            if 'intent' in transition_route:
                if transition_route.target_page in temp_page_name_list:
                    resources['intents'].append(transition_route.intent)

        return resources

    def _get_route_groups_and_intents(self, page_object, flow_id, resources):
        """Extract Intent resources from Transition Route Groups on a Page."""
        route_groups = self.route_groups.list_transition_route_groups(flow_id)

        if 'transition_route_groups' in page_object:
            for trg in page_object.transition_route_groups:
                resources['route_groups'].append(trg)
                for route_group in route_groups:
                    if trg == route_group.name:
                        for transition_route in route_group.transition_routes:
                            resources['intents'].append(
                                transition_route.intent)

        return resources

    def _remap_parameters_in_intent(
        self,
        source_agent,
        destination_agent,
        intent_object):
        """Remap the Entity Type Resource ID from the Source to Destination.

        Internal function to find the Source Entity Type Resource ID, convert
        it to the Display Name, perform a lookup in the Destination Agent, and
        convert that final Resource ID to the Desitnation Agent Entity Type
        Resource ID
        """

        source_entities_map = self.entities.get_entities_map(source_agent)
        destination_entities_map = self.entities.get_entities_map(
            destination_agent, reverse=True
            )

        for param in intent_object.parameters:
            if 'sys.' in param.entity_type:
                pass
            else:
                source_name = source_entities_map[param.entity_type]
                destination_name = destination_entities_map[source_name]
                param.entity_type = destination_name

        return intent_object

    def _update_intent_via_copy(
        self,
        intent_display_name,
        intent_object,
        destination_agent):
        """Update the Source Intent Display Name in the Destination Agent.

        Internal method to update the Destination Intent based on the defined
        Source Intent Display Name. This is specifically used when the
        Destination Agent already contains the Intent Display Name specified.
        This performs a lookup based on Intent Display Name, retrieves the
        Destination Intent object, and then updates it appropriately.
        """

        destination_intents = self.intents.list_intents(destination_agent)
        for intent in destination_intents:
            if intent.display_name == intent_display_name:
                destination_intent_obj = intent

        self.intents.update_intent(destination_intent_obj.name, intent_object)
        logging.info('Intent %s updated successfully',
          intent_object.display_name)

    def _create_intent_via_copy(
        self,
        destination_agent,
        intent_object):
        """Create the Source Intent in the Destination Agent."""
        try:
            self.intents.create_intent(destination_agent, intent_object)
            logging.info(
                'Intent %s created successfully', intent_object.display_name)

        except core_exceptions.AlreadyExists as error:
            print(error)
            print(
                'If you are trying to update an existing Intent, use the'
                + ' \'update\' option instead'
            )

    def _get_resource_objects(
        self,
        source_agent,
        resources,
        resources_objects,
        skip_list):

        if 'entities' not in skip_list:
            source_entities = self.entities.list_entity_types(source_agent)
            for entity in source_entities:
                if entity.name in resources['entities']:
                    resources_objects['entities'].append(entity)

        if 'intents' not in skip_list:
            source_intents = self.intents.list_intents(source_agent)
            for intent in source_intents:
                if intent.name in resources['intents']:
                    resources_objects['intents'].append(intent)

        if 'webhooks' not in skip_list:
            source_webhooks = self.webhooks.list_webhooks(source_agent)
            for webhook in source_webhooks:
                if webhook.name in resources['webhooks']:
                    resources_objects['webhooks'].append(webhook)

        if 'route_groups' not in skip_list:
            source_flows_map = self.flows.get_flows_map(
                source_agent, reverse=True
            )
            source_route_groups = (
                self.route_groups.list_transition_route_groups(
                    source_flows_map['Default Start Flow']
                )
            )
            for route_group in source_route_groups:
                if route_group.name in resources['route_groups']:
                    resources_objects['route_groups'].append(route_group)

        return resources_objects

    def _create_webhook_resources(
        self,
        destination_agent,
        resources_objects,
        resources_skip_list):
        for webhook in resources_objects['webhooks']:
            logging.info('Creating Webhook %s...', webhook.display_name)
            try:
                self.webhooks.create_webhook(destination_agent, webhook)
                resources_skip_list['webhooks'].append(webhook.display_name)
                logging.info(
                    'Webhook %s created successfully.', webhook.display_name)

            except core_exceptions.AlreadyExists as error:
                logging.info(error)

        return resources_skip_list

    def _create_entity_resources(
        self,
        destination_agent,
        resources_objects,
        resources_skip_list):

        for entity in resources_objects['entities']:
            logging.info('Creating Entity %s...', entity.display_name)
            try:
                self.entities.create_entity_type(destination_agent, entity)
                resources_skip_list['entities'].append(entity.display_name)
                logging.info(
                    'Entity %s created successfully.', entity.display_name)

            except core_exceptions.AlreadyExists as error:
                print(error)

        return resources_skip_list

    def _create_intent_resources(
        self,
        source_agent,
        destination_agent,
        resources_objects,
        resources_skip_list):

        for intent in resources_objects['intents']:
            logging.info('Creating Intent %s...', intent.display_name)
            time.sleep(1) # sleep to not overrun API rate limiter

            if 'parameters' in intent:
                intent = self._remap_parameters_in_intent(
                    source_agent, destination_agent, intent)
            try:
                self.intents.create_intent(destination_agent, intent)
                resources_skip_list['intents'].append(intent.display_name)
                logging.info('Intent %s created successfully',
                  intent.display_name)
            except core_exceptions.AlreadyExists as error:
                print(error)

        return resources_skip_list

    def _create_route_group_resources( #pylint: disable=too-many-arguments
        self,
        source_agent,
        destination_agent,
        destination_flow,
        resources_objects,
        resources_skip_list):

        source_flows_map = self.flows.get_flows_map(source_agent)
        source_intents_map = self.intents.get_intents_map(source_agent)
        source_webhooks_map = self.webhooks.get_webhooks_map(source_agent)
        source_pages_map = self.pages.get_pages_map(
            source_flows_map['Default Start Flow'])

        destination_flows = self.flows.get_flows_map(
            destination_agent, reverse=True)
        destination_intents_map = self.intents.get_intents_map(
                destination_agent, reverse=True)
        destination_webhooks_map = self.webhooks.get_webhooks_map(
                destination_agent, reverse=True)
        destination_pages_map = self.pages.get_pages_map(
                destination_flows[destination_flow], reverse=True
            )

        for route_group in resources_objects['route_groups']:
            logging.info(
                'Creating Route Group %s...', route_group.display_name)
            for trans_route in route_group.transition_routes:
                source_name = source_intents_map[trans_route.intent]
                destination_name = destination_intents_map[source_name]
                trans_route.intent = destination_name

                if 'trigger_fulfillment' in trans_route:
                    if 'webhook' in trans_route.trigger_fulfillment:
                        source_webhook = source_webhooks_map[
                            trans_route.trigger_fulfillment.webhook
                        ]
                        destination_webhook = destination_webhooks_map[
                            source_webhook
                        ]
                        trans_route.trigger_fulfillment.webhook = (
                            destination_webhook)

                if 'target_page' in trans_route:
                    if trans_route.target_page.split('/')[-1] == 'END_FLOW':
                        trans_route.target_page = (
                            destination_flows[destination_flow]
                            + '/pages/END_FLOW'
                        )
                    else:
                        source_page = source_pages_map[trans_route.target_page]
                        destination_page = destination_pages_map[
                            source_page
                        ]
                        trans_route.target_page = destination_page

            try:
                self.route_groups.create_transition_route_group(
                    destination_flows[destination_flow], route_group
                )
                resources_skip_list['route_groups'].append(
                    route_group.display_name)
                logging.info('Route Group %s created successfully',
                  route_group.display_name)
            except core_exceptions.AlreadyExists as error:
                print(error)

        return resources_skip_list

    def copy_intent_to_agent(
        self,
        intent_display_name: str,
        source_agent: str,
        destination_agent: str,
        copy_option: str = 'create',
    ):
        """Copy an Intent object from one CX agent to another.

        Args:
          intent_display_name: The human readable display name of the intent.
          source_agent: the Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          destination_agent: the Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          copy_optoion: The update method of the copy to the new agent.
            One of 'create' or 'update'. Defaults to 'create'
        """
        # retrieve from source agent
        intents_map = self.intents.get_intents_map(source_agent, reverse=True)
        intent_id = intents_map[intent_display_name]
        intent_object = self.intents.get_intent(intent_id)

        if 'parameters' in intent_object:
            intent_object = self._remap_parameters_in_intent(
                source_agent, destination_agent, intent_object
            )

        if copy_option == 'update':
            self._update_intent_via_copy(
                intent_display_name, intent_object,destination_agent)

        elif copy_option == 'create':
            self._create_intent_via_copy(
                destination_agent, intent_object
            )

        else:
            logging.info('Invalid option. Please use \'create\' or \'update\'')


    def copy_entity_type_to_agent(
        self,
        entity_type_display_name,
        source_agent,
        destination_agent
    ):
        """Copy an Entity Type object from one CX agent to another.

        Args:
          entity_type_display_name: The human readable display name of the
            entity.
          source_agent: the Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          destination_agent: the Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
        """
        # retrieve from source agent
        entity_map = self.entities.get_entities_map(source_agent, reverse=True)
        entity_id = entity_map[entity_type_display_name]
        entity_object = self.entities.get_entity_type(entity_id)

        # push to destination agent
        try:
            self.entities.create_entity_type(destination_agent, entity_object)
            logging.info('Entity Type %s created successfully',
              entity_object.display_name)

        except core_exceptions.AlreadyExists as error:
            print(error)

    def create_page_shells(
        self,
        pages_list: List[types.Page],
        destination_agent: str,
        destination_flow: str = 'Default Start Flow',
    ):
        """Create blank DFCX Page object(s) with given Display Name.

        This function aids in the copy/pasting of pages from one DFCX agent to
        another by first creating blank Page "shells" in the destination agent
        using the human-readable display names. These Pages will then be
        retrieved by Page ID to use in the final copy/paste of the Page object
        from source to destination.

        Args:
          pages_list: List of Page objects to extract page names from
          destination_agent: DFCX Agent ID of the Destination Agent
          destination_flow: DFCX Flow ID of the Destination Flow. If no Flow ID
            is provided, Default Start Flow will be used.

        Returns:
          None
        """
        destination_flows = self.flows.get_flows_map(
            destination_agent, reverse=True
        )

        for page in pages_list:
            try:
                self.pages.create_page(
                    destination_flows[destination_flow],
                    display_name=page.display_name,
                )
                logging.info(
                    'Page %s created successfully', page.display_name
                )
            except core_exceptions.AlreadyExists as error:
                logging.info(error)
                continue

    def copy_paste_agent_resources( #pylint: disable=too-many-arguments
        self,
        resources: Dict[str, str],
        source_agent: str,
        destination_agent: str,
        destination_flow: str = 'Default Start Flow',
        skip_list: List[str] = None
    ):
        """Copy/Paste Agent level resources from one DFCX agent to another.

        Agent level resources in DFCX are resources like Entities, Intents, and
        Webhooks which are not Flow dependent. This method allows the user to
        provide a dictionary of Agent Resources and Resources IDs to be copied
        from a Source agent to a Destination agent. *NOTE* That this method
        will also copy all Route Groups from Default Start Flow only.

        To obtain the resource_dict in the proper format, you can use the
        get_page_dependencies() method included in the CopyUtil Class.

        Args:
          resource_dict: Dictionary of Lists of DFCX Resource IDs with keys
            corresponding to the Resource type (i.e. intents, entities, etc.)
            and values corresponding to the Resource ID itself.
          source_agent: DFCX Source Agent ID (Name)
          destination_agent: DFCX Destination Agent ID (Name)
          destination_flow: (Optional) Defaults to 'Default Start Flow'
          skip_list: (Optional) List of resources to exclude. Use the following
              strings: 'intents', 'entities', 'webhooks', 'route_groups'

        Returns:
          A dictionary with possible keys being webhooks, entities, intents,
          and route_groups, with keys missing if they were in the skip_list.
          Each value is a list of display names of created CX resources.
        """
        resources_objects = defaultdict(list)
        resources_skip_list = defaultdict(list)

        resources_objects = self._get_resource_objects(
            source_agent, resources, resources_objects, skip_list)

        # Create Objects in Destination Agent
        # For all objects, we will attempt to create them in the destination
        # Agent. If the Resource is a duplicate, then we will skip it and add
        # it to the resources_skip_list. Duplicates are determined by
        # display_name only at this time.

        if 'webhooks' in resources_objects and 'webhooks' not in skip_list:
            resources_skip_list = self._create_webhook_resources(
                destination_agent, resources_objects, resources_skip_list)

        if 'entities' in resources_objects and 'entities' not in skip_list:
            resources_skip_list = self._create_entity_resources(
                destination_agent, resources_objects, resources_skip_list)

        if 'intents' in resources_objects and 'intents' not in skip_list:
            resources_skip_list = self._create_intent_resources(
                source_agent, destination_agent, resources_objects,
                resources_skip_list)

        if (
            'route_groups' in resources_objects and
            'route_groups' not in skip_list):
            resources_skip_list = self._create_route_group_resources(
                source_agent, destination_agent, destination_flow,
                resources_objects,resources_skip_list)

        return resources_skip_list

    def convert_from_source_page_dependencies(
        self,
        agent_id: str,
        pages_list: List[types.Page],
        flow: str = 'Default Start Flow',
    ) -> List[types.Page]:
        """Convert all Source Agent dependencies to Display Names.

        In order to copy resources over to a Destination Agent, we need to
        first convert all of the Resource IDs to their respective Display
        Names. We will then use the Display Names to convert back to the
        Destination Agent resource IDs using another method.

        Args:
          agent_id: the source Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          pages_list: A list of DFCX Page objects
          flow: The flow display name. Defaults to "Default Start Flow"

        Returns:
          The modified list of pages, with source agent resource IDs converted
          to display names.
        """

        pages_mod = copy.deepcopy(pages_list)

        intents_map = self.intents.get_intents_map(agent_id)
        entities_map = self.entities.get_entities_map(agent_id)
        webhooks_map = self.webhooks.get_webhooks_map(agent_id)
        flows_map = self.flows.get_flows_map(agent_id, reverse=True)
        pages_map = self.pages.get_pages_map(flows_map[flow])
        rgs_map = self.route_groups.get_route_groups_map(flows_map[flow])

        # For each page, recurse through the resources and look for
        # specific resource types that will have local agent
        # dependencies, then replace that UUID with a literal
        # str display_name. We will use this display_name to map back
        # to the appropriate destination UUID later.
        for page in pages_mod:

            if 'entry_fulfillment' in page:
                page = self._convert_entry_webhooks(page, webhooks_map)

            if 'transition_routes' in page:
                page = self._convert_trans_routes(
                    page, pages_map, intents_map, webhooks_map,
                    convert_type='source')

            if 'event_handlers' in page:
                page = self._convert_event_handlers(
                    page, pages_map, webhooks_map, convert_type='source')

            if 'form' in page:
                if 'parameters' in page.form:
                    page = self._convert_form_parameters(
                        page, pages_map, webhooks_map, entities_map,
                        convert_type='source')

            if 'transition_route_groups' in page:
                temp_list = []
                for trg in page.transition_route_groups:
                    temp_list.append(rgs_map[trg])

                    page.transition_route_groups = temp_list

        return pages_mod

    def convert_to_destination_page_dependencies(
        self,
        agent_id: str,
        pages_list: List[types.Page],
        flow: str = 'Default Start Flow'
    ) -> List[types.Page]:
        """Convert from Display Names to Destination Agent Resource IDs.

        In order to copy resources over to a Destination Agent, we need to
        look up all the Display Names we previously converted and remap them
        to their respective Destination Agent Resource IDs. We will do this
        by extracting maps of the Destination Agent Resources and then
        performing a lookup with the Page objects in our pages_list.

        Args:
          agent_id: the target Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          pages_list: A list of DFCX Page objects
          flow: The flow display name. Defaults to "Default Start Flow"

        Returns:
          The modified list of pages, with display names converted to
          resource IDs for the target agent.
        """

        pages_mod = copy.deepcopy(pages_list)

        intents_map = self.intents.get_intents_map(agent_id, reverse=True)
        entities_map = self.entities.get_entities_map(agent_id, reverse=True)
        webhooks_map = self.webhooks.get_webhooks_map(agent_id, reverse=True)
        flows_map = self.flows.get_flows_map(agent_id, reverse=True)
        pages_map = self.pages.get_pages_map(flows_map[flow], reverse=True)
        rgs_map = self.route_groups.get_route_groups_map(
            flows_map[flow], reverse=True
        )

        # For each page, recurse through the resources and look for
        # specific resource types that have been replaced with literal
        # string display_name. Perform a lookup using the map resources
        # and replace the str display_name with the appropriate str UUID

        for page in pages_mod:
            page.name = pages_map[page.display_name]

            if 'entry_fulfillment' in page:
                page = self._convert_entry_webhooks(page, webhooks_map)

            if 'transition_routes' in page:
                page = self._convert_trans_routes(
                    page, pages_map, intents_map, webhooks_map,
                    convert_type='destination'
                )

            if 'event_handlers' in page:
                page = self._convert_event_handlers(
                    page, pages_map, webhooks_map, convert_type='destination')


            if 'form' in page:
                if 'parameters' in page.form:
                    page = self._convert_form_parameters(
                        page, pages_map, webhooks_map, entities_map,
                        convert_type='destination')

            if 'transition_route_groups' in page:
                temp_list = []
                for trg in page.transition_route_groups:
                    temp_list.append(rgs_map[trg])

                page.transition_route_groups = temp_list

        return pages_mod

    def convert_start_page_dependencies(
        self,
        agent_id,
        start_page,
        agent_type='source',
        flow='Default Start Flow',
    ):
        """Convert all Source Agent Start page dependencies to Display Names.

        In order to copy resources over to a Destination Agent, we need to
        first convert all of the Resource IDs to their respective Display
        Names. We will then use the Display Names to convert back to the
        Destination Agent resource IDs using another method. Start Pages are a
        special type of Page that exists inside of the Flow object, so they
        have to be handled differently.

        Args:
          agent_id: the source Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          start_page: The DFCX Flow object corresponding to the start page of
            the flow to be copied
          agent_type: "source" or "destination". Defaults to "source"
          flow: The flow display name. Defaults to "Default Start Flow"

        Returns:
          The modified start page, with the resource IDs converted to
          display names for the source option and the display names converted
          to resource IDs for the destination option.
        """
        page_mod = copy.deepcopy(start_page)

        if agent_type == 'source':
            intents_map = self.intents.get_intents_map(agent_id)
            webhooks_map = self.webhooks.get_webhooks_map(agent_id)
            flows_map = self.flows.get_flows_map(agent_id, reverse=True)
            pages_map = self.pages.get_pages_map(flows_map[flow])

            for trans_route in page_mod.transition_routes:
                if 'target_page' in trans_route:
                    if trans_route.target_page.split('/')[-1] == 'END_FLOW':
                        trans_route.target_page = 'END_FLOW'
                    elif trans_route.target_page.split('/')[-1] == 'START_PAGE':
                        trans_route.target_page = 'START_PAGE'
                    else:
                        trans_route.target_page = pages_map[
                            trans_route.target_page]

                if 'intent' in trans_route:
                    trans_route.intent = intents_map[trans_route.intent]
                    if 'webhook' in trans_route.trigger_fulfillment:
                        trans_route.trigger_fulfillment.webhook = webhooks_map[
                            trans_route.trigger_fulfillment.webhook
                        ]
                elif ('condition' in trans_route and
                'webhook' in trans_route.trigger_fulfillment):
                    trans_route.trigger_fulfillment.webhook = webhooks_map[
                        trans_route.trigger_fulfillment.webhook
                    ]

        elif agent_type == 'destination':
            final_trs = []
            intents_map = self.intents.get_intents_map(agent_id, reverse=True)
            webhooks_map = self.webhooks.get_webhooks_map(
                agent_id, reverse=True
            )
            flows_map = self.flows.get_flows_map(agent_id, reverse=True)
            pages_map = self.pages.get_pages_map(flows_map[flow], reverse=True)

            page_mod.name = flows_map[flow]
            print(page_mod.name)

            for trans_route in page_mod.transition_routes:
                if 'target_page' in trans_route:
                    if trans_route.target_page in ['END_FLOW', 'START_PAGE']:
                        if trans_route.target_page == 'END_FLOW':
                            trans_route.target_page = (
                                flows_map[flow] + '/pages/END_FLOW')
                        elif trans_route.target_page == 'START_PAGE':
                            trans_route.target_page = (
                                flows_map[flow] + '/pages/START_PAGE'
                            )

                    elif trans_route.target_page in pages_map:
                        trans_route.target_page = pages_map[
                            trans_route.target_page]

                if 'intent' in trans_route:
                    if trans_route.intent not in intents_map:
                        logging.info(
                            'Intent %s not in Intents Map. Skipping.',
                            trans_route.intent)
                    elif trans_route.intent in intents_map:
                        trans_route.intent = intents_map[trans_route.intent]
                        if 'webhook' in trans_route.trigger_fulfillment:
                            trans_route.trigger_fulfillment.webhook = (
                                webhooks_map[
                                    trans_route.trigger_fulfillment.webhook
                                    ]
                            )

                        final_trs.append(trans_route)

                elif ('condition' in trans_route and
                'webhook' in trans_route.trigger_fulfillment):
                    trans_route.trigger_fulfillment.webhook = webhooks_map[
                        trans_route.trigger_fulfillment.webhook
                    ]
                    final_trs.append(trans_route)

            page_mod.transition_routes = final_trs

        return page_mod


    def get_page_dependencies(self, obj_list):
        """Pass in DFCX Page object(s) and retrieve all resource dependencies.

        Args:
          obj_list: a List of one or more DFCX Page Objects

        Returns:
          Dictionary containing all of the resource objects
        """
        resources = defaultdict(list)
        flow_id = '/'.join(obj_list[0].name.split('/')[0:8])

        # Loop through Pages and find all dependencies
        for page in obj_list:
            resources = self._get_entry_webhooks(page, resources)
            resources = self._get_condition_route_webhooks(page, resources)
            resources = self._get_intent_route_intents(page, resources)
            resources = self._get_form_entity_types(page, resources)
            resources = self._get_route_groups_and_intents(page, flow_id,
              resources)

        # Start Pages of Flows are special Page-like objects and have different
        # structure, so we need to extract the resources from them differently.
        resources = self._get_flow_intent_route_intents(
            flow_id, obj_list, resources)

        # Final check to look for additional Entity dependencies
        if 'intents' in resources:
            resources = self._get_intent_entity_dependencies(resources)

        for key in resources:
            resources[key] = set(resources[key])

        return resources



"""DFCX End to End Conversation Functions"""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import traceback
import uuid

from typing import Dict
from operator import attrgetter
from threading import Thread

import pandas as pd

from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.api_core import exceptions as core_exceptions
from proto.marshal.collections import repeated
from proto.marshal.collections import maps

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import flows
from dfcx_scrapi.core import pages

logging.basicConfig(
    format="[dfcx] %(levelname)s:%(message)s", level=logging.INFO
)

MAX_RETRIES = 3


class DialogflowConversation(scrapi_base.ScrapiBase):
    """Class that wraps the SessionsClient to hold end to end conversations
    and maintain internal session state
    """

    def __init__(
        self,
        config=None,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        agent_id: str = None,
        language_code: str = "en",
    ):

        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            agent_id=agent_id,
        )

        logging.info(
            f"create conversation with creds_path: \
                {creds_path} | agent_id: {agent_id}",
        )

        if agent_id or config["agent_path"]:
            self.agent_id = agent_id or config["agent_path"]

        self.language_code = language_code or config["language_code"]

        self.start_time = None
        self.query_result = None
        self.session_id = None
        self.turn_count = None
        self.agent_env = {}  # empty
        self.restart()
        self.flows = flows.Flows(creds=self.creds)
        self.pages = pages.Pages(creds=self.creds)

    @staticmethod
    def _get_match_type_from_map(match_type: int):
        """Translates the match_type enum int value into a more descriptive
        string.
        """
        match_type_map = {
            0: "MATCH_TYPE_UNSPECIFIED",
            1: "INTENT",
            2: "DIRECT_INTENT",
            3: "PARAMETER_FILLING",
            4: "NO_MATCH",
            5: "NO_INPUT",
            6: "EVENT",
        }

        return match_type_map[match_type]

    @staticmethod
    def _validate_test_set_input(test_set: pd.DataFrame):
        """Validates that all pages referenced in the test set exist in the
        agent.
        """
        mask = test_set.page_id.isna().to_list()
        invalid_pages = set(test_set.page_display_name[mask].to_list())

        if invalid_pages:
            raise Exception(
                "The following Pages are invalid and missing Page "
                f"IDs: \n{invalid_pages}\n\nPlease ensure that your Page "
                "Display Names do not contain typos.\nFor Default Start Page "
                "use the special page display name START_PAGE."
            )

    @staticmethod
    def progress_bar(current, total, bar_length=50, type_="Progress"):
        """Display progress bar for processing."""
        percent = float(current) * 100 / total
        arrow = "-" * int(percent / 100 * bar_length - 1) + ">"
        spaces = " " * (bar_length - len(arrow))
        print(
            f"{type_}({current}/{total})" + f"[{arrow}{spaces}] {percent:.2f}%",
            end="\r",
        )

    @staticmethod
    def _build_query_params_object(parameters, current_page, disable_webhook):
        if parameters:
            query_params = types.session.QueryParameters(
                disable_webhook=disable_webhook,
                parameters=parameters,
                current_page=current_page,
            )
        else:
            query_params = types.session.QueryParameters(
                disable_webhook=disable_webhook, current_page=current_page
            )

        return query_params

    @staticmethod
    def _build_query_input_object(input_obj, language_code):
        if "dtmf" in input_obj:
            digits = str(input_obj["dtmf"])

            finish_digit = None
            if "finish_digit" in input_obj:
                finish_digit = str(input_obj["finish_digit"])

            dtmf_input = types.session.DtmfInput(
                digits=digits, finish_digit=finish_digit
            )
            query_input = types.session.QueryInput(
                dtmf=dtmf_input,
                language_code=language_code,
            )

        elif "intent" in input_obj:
            intent_input = types.session.IntentInput(intent=input_obj["intent"])
            query_input = types.session.QueryInput(
                intent=intent_input, language_code=language_code
            )

        elif "event" in input_obj:
            event_input = types.session.EventInput(event=input_obj["event"])
            query_input = types.session.QueryInput(
                event=event_input, language_code=language_code
            )

        elif "text" in input_obj:
            text = input_obj["text"]
            logging.debug("Input text: %s", text)
            text_input = types.session.TextInput(text=text)
            query_input = types.session.QueryInput(
                text=text_input,
                language_code=language_code,
            )

        return query_input

    @staticmethod
    def _gather_text_responses(text_message):

        flat_texts = "\n".join(text_message.text)

        return flat_texts

    def _gather_response_messages(self, response_messages):
        rm_gathered = []
        for msg in response_messages:
            if msg.payload:
                msg = {
                    "payload": self.recurse_proto_marshal_to_dict(msg.payload)
                }

            elif msg.play_audio:
                msg = {"play_audio": {"audio_uri": msg.play_audio.audio_uri}}

            elif msg.live_agent_handoff:
                msg = {
                    "live_agent_handoff": self.recurse_proto_marshal_to_dict(
                        msg.live_agent_handoff.metadata
                    )
                }

            elif msg.conversation_success:
                msg = {
                    "conversation_success": self.recurse_proto_marshal_to_dict(
                        msg.conversation_success.metadata
                    )
                }

            elif msg.output_audio_text:
                msg = {"output_audio_text": msg.output_audio_text.text}

            elif msg.text:
                msg = {"text": self._gather_text_responses(msg.text)}

            rm_gathered.append(msg)

        return rm_gathered

    def _gather_query_result_parameters(self, input_parameters):
        output_parameters = {}
        for param in input_parameters:
            val = input_parameters[param]

            # If we find a RepeatedComposite (i.e. List) we will recurse
            # down and convert to lists/dics/str as needed.
            if isinstance(val, repeated.RepeatedComposite):
                val = self.recurse_proto_repeated_composite(val)

            elif isinstance(val, maps.MapComposite):
                val = self.recurse_proto_marshal_to_dict(val)

            output_parameters[param] = val

        return output_parameters

    def _page_id_mapper(self):
        """Initializes the agent_pages_map dataframe.

        This dataframe contains the flow_display_name, page_display_name,
        and page_id for each page in the agent.
        """
        agent_pages_map = pd.DataFrame()
        flow_map = self.flows.get_flows_map(agent_id=self.agent_id)
        for flow_id in flow_map.keys():

            page_map = self.pages.get_pages_map(flow_id=flow_id)

            flow_mapped = pd.DataFrame.from_dict(page_map, orient="index")
            flow_mapped["page_id"] = flow_mapped.index

            flow_mapped = flow_mapped.rename(columns={0: "page_display_name"})

            # add start page
            start_page_id = flow_id + "/pages/START_PAGE"
            flow_mapped = pd.concat(
                [
                    flow_mapped,
                    pd.DataFrame(
                        columns=["page_display_name", "page_id"],
                        data=[["START_PAGE", start_page_id]],
                    ),
                ]
            )

            flow_mapped.insert(0, "flow_display_name", flow_map[flow_id])
            agent_pages_map = pd.concat([agent_pages_map, flow_mapped])

        self.agent_pages_map = agent_pages_map.reset_index(drop=True)

    def _get_reply_results(self, utterance, page_id, results, i):
        """Get results of single text utterance to CX Agent.

        Args:
          utterance: Text to send to the bot for testing.
          page_id: Specified CX Page to send the utterance request to
          results: Pandas Dataframe to capture and store the results
          i: Internal tracking for Python Threading
        """
        response = self.reply(
            send_obj={"text": utterance}, current_page=page_id, restart=True
        )

        target_page = response["page_name"]

        results["target_page"][i] = target_page
        results["match"][i] = response["match"]

    def _get_intent_detection(self, test_set: pd.DataFrame):
        """Gets the results of a subset of Intent Detection tests.

        NOTE - This is an internal method used by run_intent_detection to
        manage parallel intent detection requests and should not be used as a
        standalone function.
        """

        self._page_id_mapper()
        test_set_mapped = pd.merge(
            test_set,
            self.agent_pages_map,
            on=["flow_display_name", "page_display_name"],
            how="left",
        )
        utterances = list(test_set_mapped["utterance"])
        page_ids = list(test_set_mapped["page_id"])

        self._validate_test_set_input(test_set_mapped)

        threads = [None] * len(utterances)
        results = {
            "target_page": [None] * len(utterances),
            "match":[None] * len(utterances),
        }
        for i, (utterance, page_id) in enumerate(zip(utterances, page_ids)):
            threads[i] = Thread(
                target=self._get_reply_results,
                args=(utterance, page_id, results, i),
            )
            threads[i].start()

        for idx, _ in enumerate(threads):
            threads[idx].join()

        test_set_mapped["target_page"] = results["target_page"]
        test_set_mapped["match"] = results["match"]
        test_set_mapped = test_set_mapped.drop(columns=["page_id"])
        intent_detection = test_set_mapped.copy()

        return intent_detection


    def restart(self):
        """Starts a new session/conversation for this agent"""
        self.session_id = uuid.uuid4()
        self.turn_count = 0

    def set_agent_env(self, param, value):
        """Setting changes related to the environment"""
        logging.info("setting agent_env param:[%s] = value:[%s]", param, value)
        self.agent_env[param] = value

    def checkpoint(self, msg=None, start=False):
        """Print a checkpoint to time progress and debug bottleneck"""
        if start:
            start_time = time.perf_counter()
            self.start_time = start_time
        else:
            start_time = self.start_time
        duration = round((time.perf_counter() - start_time), 2)
        if duration > 2:
            if msg:
                print(f"{duration:0.2f}s {msg}")

    def reply(
        self,
        send_obj: Dict[str, str],
        restart: bool = False,
        retries: int = 0,
        current_page: str = None,
        checkpoints: bool = False,
    ):
        """Runs intent detection on one utterance and gets the agent reply.

        Args:
          send_obj: Dictionary with the following structure:
            {'text': str,
            'params': Dict[str,str],
            'dtmf': str}
          restart: Boolean flag that determines whether to use the existing
            session ID or start a new conversation with a new session ID.
            Passing True will create a new session ID on subsequent calls.
            Defaults to False.
          retries: used for recurse calling this func if API fails
          current_page: Specify the page id to start the conversation from
          checkpoints: Boolean flag to enable/disable Checkpoint timer
            debugging. Defaults to False.

        Returns:
          A dictionary for the agent reply to to the submitted text.
            Includes keys response_messages, confidence, page_name,
            intent_name, match_type, match, other_intents, and params.
        """
        text = send_obj.get("text")
        send_params = send_obj.get("params")

        if not text:
            logging.warning(f"Input Text is empty. {send_obj}")

        if text and len(text) > 256:
            logging.warning(
                "Text input is too long. Truncating to 256 characters."
            )
            text = text[0:256]
            logging.warning(f"TRUNCATED TEXT: {text}")

        custom_environment = self.agent_env.get("environment")
        disable_webhook = self.agent_env.get("disable_webhook") or False

        if checkpoints:
            self.checkpoint(start=True)

        if restart:
            self.restart()

        client_options = self._set_region(self.agent_id)
        session_client = services.sessions.SessionsClient(
            credentials=self.creds, client_options=client_options
        )
        session_path = f"{self.agent_id}/sessions/{self.session_id}"

        if custom_environment:
            logging.info("req using env: %s", custom_environment)
            session_path = (
                f"{self.agent_id}/environments/"
                f"{custom_environment}/sessions/{self.session_id}"
            )

        # Build Query Params object
        query_params = self._build_query_params_object(
            send_params, current_page, disable_webhook
        )

        # Build Query Input object
        query_input = self._build_query_input_object(
            send_obj, self.language_code
        )

        request = types.session.DetectIntentRequest(
            session=session_path,
            query_input=query_input,
            query_params=query_params,
        )

        logging.debug("query_params: %s", query_params)
        logging.debug("request %s", request)

        response = None
        try:
            response = session_client.detect_intent(request=request)

        except core_exceptions.InternalServerError as err:
            logging.error(
                "---- ERROR --- InternalServerError caught on CX.detect %s", err
            )
            logging.error("text: %s", text)
            logging.error("query_params: %s", query_params)
            logging.error("query_input: %s", query_input)
            return {}

        except core_exceptions.ClientError as err:
            logging.error(
                "---- ERROR ---- ClientError caught on CX.detect %s", err
            )
            template = "An exception of type {0} occurred. \nArguments:\n{1!r}"
            message = template.format(type(err).__name__, err.args)
            logging.error("err name %s", message)

            logging.error("text %s", text)
            logging.error("query_params %s", query_params)
            logging.error("query_input %s", query_input)
            logging.error(traceback.print_exc())
            retries += 1
            if retries < MAX_RETRIES:
                logging.error("retrying")
                return self.reply(send_obj, restart=restart, retries=retries)
            else:
                logging.error("MAX_RETRIES exceeded")
                return {}

        if checkpoints:
            self.checkpoint("<< got response")
        query_result = response.query_result
        logging.debug("dfcx>qr %s", query_result)
        self.query_result = query_result
        reply = {}

        # Gather Response Messages into List of Dicts
        if query_result.response_messages:
            response_messages = self._gather_response_messages(
                query_result.response_messages
            )
        else:
            response_messages = None

        # Convert params structures from Proto to standard python data types
        if query_result.parameters:
            params = self._gather_query_result_parameters(
                query_result.parameters
            )
        else:
            params = None

        reply["response_messages"] = response_messages
        reply["confidence"] = query_result.intent_detection_confidence
        reply["page_name"] = query_result.current_page.display_name
        reply["intent_name"] = query_result.intent.display_name
        reply["match_type"] = self._get_match_type_from_map(
            query_result.match.match_type
        )
        reply["match"] = query_result.match
        reply["other_intents"] = self.format_other_intents(query_result)
        reply["params"] = params

        logging.debug("reply %s", reply)

        return reply

    def format_other_intents(self, query_result):
        """Unwind protobufs into more friendly dict"""
        other_intents = query_result.diagnostic_info.get(
            "Alternative Matched Intents"
        )
        items = []
        rank = 0
        for alt in other_intents:
            items.append(
                {
                    "name": alt.get("DisplayName"),
                    "score": alt.get("Score"),
                    "rank": rank,
                }
            )
            rank += 1

        if self:
            return items

        return None

    def getpath(self, obj, xpath, default=None):
        """Get data at a pathed location out of object internals"""
        elem = obj
        try:
            for xpitem in xpath.strip("/").split("/"):
                try:
                    xpitem = int(xpitem)
                    elem = elem[xpitem]  # dict
                except ValueError:
                    elem = elem.get(xpitem)  # array
        except KeyError:
            logging.warning("failed to getpath: %s ", xpath)
            return default

        logging.info("OK getpath: %s", xpath)
        if self:
            return elem

        return None

    def run_intent_detection(
        self,
        test_set: pd.DataFrame,
        chunk_size: int = 300,
        rate_limit: float = 20,
    ):
        """Tests a set of utterances for intent detection against a CX Agent.

        This function uses Python Threading to run tests in parallel to
        expedite intent detection testing for Dialogflow CX agents. The default
        quota for Text requests/min is 1200. Ref:
          https://cloud.google.com/dialogflow/quotas#table

        Args:
          test_set: A Pandas DataFrame with the following schema.
            flow_display_name: str
            page_display_name: str
              - NOTE, when using the Default Start Page of a Flow you must
                define it as the special display name START_PAGE
            utterance: str
          chunk_size: Determines the number of text requests to send in
            parallel. This should be adjusted based on your test_set size and
            the Quota limits set for your GCP project. Default is 300.
          rate_limit: Number of seconds to wait between running test set chunks

        Returns:
          A Pandas DataFrame consisting of the original
            DataFrame plus an additional column for the detected intent with
            the following schema.
              flow_display_name: str
              page_display_name: str
              utterance: str
              detected_intent: str
              confidence: float
              target_page: str
        """

        result = pd.DataFrame()
        for start in range(0, test_set.shape[0], chunk_size):
            test_set_chunk = test_set.iloc[start : start + chunk_size]
            result_chunk = self._get_intent_detection(test_set=test_set_chunk)
            result = pd.concat([result, result_chunk])
            self.progress_bar(start, test_set.shape[0])
            time.sleep(rate_limit)
        self.progress_bar(test_set.shape[0], test_set.shape[0])

        result = self._unpack_match(result)
        return result

    def _unpack_match(self, df: pd.DataFrame):
        """ Unpacks a 'match' column into four component columns.

        Args:
          df: dataframe containing a column named match of types.Match

        Returns:
          A copy of df with columns match_type, confidence, parameters_set,
            and detected_intent instead of match.
        """
        df = (
            df
            .copy()
            .assign(
                match_type = lambda df: df.match.apply(
                    attrgetter("match_type._name_")),
                confidence = lambda df: df.match.apply(
                    attrgetter("confidence")),
                parameters_set = lambda df: df.match.apply(
                    attrgetter("parameters")),
                detected_intent = lambda df: df.match.apply(
                    attrgetter("intent.display_name"))
            )
            .assign(
                parameters_set = lambda df: df.parameters_set.apply(
                    lambda p: self.recurse_proto_marshal_to_dict(
                        p) if p else "")
            )
            .drop(columns="match")
        )
        return df



"""Flow Resource functions."""

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2
from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Flows(scrapi_base.ScrapiBase):
    """Core Class for CX Flow Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        flow_id: str = None,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if flow_id:
            self.flow_id = flow_id

        self.agent_id = agent_id

    def get_flows_map(
        self,
        agent_id: str,
        reverse=False
    ):
        """Exports Agent Flow Names and UUIDs into a user friendly dict.

        Args:
          agent_id: the formatted CX Agent ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing flow UUIDs as keys and display names as values
        """

        if reverse:
            flows_dict = {
                flow.display_name: flow.name
                for flow in self.list_flows(agent_id=agent_id)
            }

        else:
            flows_dict = {
                flow.name: flow.display_name
                for flow in self.list_flows(agent_id=agent_id)
            }

        return flows_dict

    def train_flow(self, flow_id: str) -> str:
        """Trains the specified flow.

        Args:
          flow_id: CX flow ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>/
              flows/<FLOW ID>

        Returns:
          A Long Running Operation (LRO) ID that can be used to
            check the status of the export using
              dfcx_scrapi.core.operations->get_lro()
        """

        request = types.flow.TrainFlowRequest()
        request.name = flow_id

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options)

        response = client.train_flow(request)

        return response

    def list_flows(
        self,
        agent_id: str
    ) -> List[types.Flow]:
        """Get a List of all Flows in the current Agent.

        Args:
          agent_id: CX Agent ID string in the proper format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>

        Returns:
          List of Flow objects
        """

        request = types.flow.ListFlowsRequest()
        request.parent = agent_id

        client_options = self._set_region(agent_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_flows(request)

        flows = []
        for page in response.pages:
            for flow in page.flows:
                flows.append(flow)
        return flows

    def get_flow_by_display_name(
        self,
        display_name: str,
        agent_id: str
    ) -> types.Flow:
        """Get a single CX Flow object based on its display name.

        Args:
          display_name: The display name of the desired Flow.
          agent_id: CX Agent ID in which the flow exists.

        Returns:
          A single CX Flow object
        """

        flows_map = self.get_flows_map(
            agent_id = agent_id,
            reverse = True
        )

        if display_name in flows_map:
            flow_id = flows_map[display_name]
        else:
            raise ValueError(
                f"Flow \"{display_name}\" " \
                f"does not exist in the specified agent."
            )

        flow = self.get_flow(flow_id=flow_id)

        return flow

    def get_flow(self, flow_id: str) -> types.Flow:
        """Get a single CX Flow object.

        Args:
          flow_id: CX Flow ID in the proper format

        Returns:
          A single CX Flow object
        """

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_flow(name=flow_id)

        return response

    def update_flow(
        self,
        flow_id: str,
        obj: types.Flow = None,
        **kwargs
    ) -> types.Flow:
        """Update a single specific CX Flow object.

        Args:
          flow_id: CX Flow ID in the proper format
          obj: (Optional) a single CX Flow object of types.Flow

        Returns:
          A copy of the updated Flow object
        """

        if obj:
            flow = obj
            flow.name = flow_id
        else:
            flow = self.get_flow(flow_id)

        # set flow attributes to args
        for key, value in kwargs.items():
            setattr(flow, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_flow(flow=flow, update_mask=mask)

        return response

    def update_nlu_settings(self, flow_id:str, **kwargs):
        """Updates flow to new NLU setting.

        Args:
          flow_id: flow id to update nlu settings for.
            model_type: (Optional) [0:unspecified, 1:MODEL_TYPE_STANDARD,
              2:Custom, 3:Advanced]
            classification_threshold: (Optional) threshold for the flow
            model_training_mode: (Optional) [0:unspecified, 1:automatic,
              2:'manual]
        """

        flow = self.get_flow(flow_id)
        current_settings = flow.nlu_settings
        for key, value in kwargs.items():
            setattr(current_settings, key, value)
        self.update_flow(flow_id=flow_id, nlu_settings=current_settings)

    def export_flow(
        self,
        flow_id: str,
        gcs_path: str,
        ref_flows: bool = True
    ) -> Dict[str, str]:
        """Exports DFCX Flow(s) into GCS bucket.

        Args:
          flow_id: the formatted CX Flow ID to export
          gcs_path: The `Google Cloud Storage URI to export the flow to. The
            format of this URI must be ``gs://<bucket-name>/<object-name>``. If
            left unspecified, the serialized flow is returned inline.
          ref_flows: Whether to export flows referenced by the specified flow.

        Returns:
          A Long Running Operation result. If successful the LRO result will
            return the Google Cloud Storage URI from the Export Flow request.
            Otherwise, it will return the corresponding error.
        """
        request = types.flow.ExportFlowRequest()
        request.name = flow_id
        request.include_referenced_flows = ref_flows
        request.flow_uri = gcs_path

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_flow(request)

        return response.result()

    def export_flow_inline(
        self,
        flow_id: str,
        ref_flows: bool = True
    ) -> bytes:
        """Export a Flow, returning uncompressed raw byte content for flow.

        Args:
          flow_id: the formatted CX Flow ID to export
          ref_flows: Whether to export flows referenced by the specified flow.

        Returns:
          Bytes representing the content of the flow.
        """
        request = types.flow.ExportFlowRequest()
        request.name = flow_id
        request.include_referenced_flows = ref_flows


        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_flow(request)

        return (response.result()).flow_content

    def import_flow(
        self,
        agent_id: str,
        gcs_path: str = None,
        flow_content: bytes = None,
        import_option: str = "KEEP"
    ) -> Dict[str, str]:
        """Imports a DFCX Flow to CX Agent. Flow can be imported from a
        GCS bucket or from raw bytes.

        Args:
          agent_id: the CX Agent ID to import the flow into.
          gcs_path: the `Google Cloud Storage URI to import flow from.
            the format of this URI must be ``gs://<bucket-name>/<object-name>``.
          flow_content: uncompressed raw byte content for flow.
          import_option: one of 'FALLBACK' or 'KEEP'. Defaults to 'KEEP'

        Returns:
          A Long Running Operation result. If successful the LRO result will
            return the Flow ID of the newly imported Flow.
            Otherwise, it will return the corresponding error.
        """

        if gcs_path and flow_content:
            raise ValueError(
                "gcs_path or flow_content required (But not both!)."
            )

        if not gcs_path and not flow_content:
            raise ValueError(
                "gcs_path or flow_content required (But not both!)."
            )

        request = types.flow.ImportFlowRequest()
        request.parent = agent_id
        request.flow_uri = gcs_path
        request.flow_content = flow_content
        request.import_option = import_option

        client_options = self._set_region(agent_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.import_flow(request)

        return response

    def delete_flow(
        self,
        flow_id: str,
        force: bool = False
    ):
        """Deletes a single CX Flow Object resource.

        Args:
          flow_id: flow to delete
          force: False means a flow will not be deleted if a route to the flow
            exists, True means the flow will be deleted as well as all the
            transition routes leading to the flow.
        """

        request = types.DeleteFlowRequest()
        request.name = flow_id
        request.force = force

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )

        client.delete_flow(request)
